# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from skytorch.client.models.io_k8s_api_core_v1_env_var import IoK8sApiCoreV1EnvVar
from skytorch.client.models.io_k8s_api_core_v1_volume_mount import IoK8sApiCoreV1VolumeMount
from typing import Optional, Set
from typing_extensions import Self

class ComputeV1alpha1ContainerOverride(BaseModel):
    """
    ContainerOverride represents parameters that can be overridden using PodSpecOverrides.
    """ # noqa: E501
    env: Optional[List[IoK8sApiCoreV1EnvVar]] = Field(default=None, description="env is the list of environment variables to set in the container.")
    name: StrictStr = Field(description="name for the container.")
    volume_mounts: Optional[List[IoK8sApiCoreV1VolumeMount]] = Field(default=None, description="volumeMounts are the volumes to mount into the container's filesystem.", alias="volumeMounts")
    __properties: ClassVar[List[str]] = ["env", "name", "volumeMounts"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of ComputeV1alpha1ContainerOverride from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in env (list)
        _items = []
        if self.env:
            for _item_env in self.env:
                if _item_env:
                    _items.append(_item_env.to_dict())
            _dict['env'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in volume_mounts (list)
        _items = []
        if self.volume_mounts:
            for _item_volume_mounts in self.volume_mounts:
                if _item_volume_mounts:
                    _items.append(_item_volume_mounts.to_dict())
            _dict['volumeMounts'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of ComputeV1alpha1ContainerOverride from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "env": [IoK8sApiCoreV1EnvVar.from_dict(_item) for _item in obj["env"]] if obj.get("env") is not None else None,
            "name": obj.get("name"),
            "volumeMounts": [IoK8sApiCoreV1VolumeMount.from_dict(_item) for _item in obj["volumeMounts"]] if obj.get("volumeMounts") is not None else None
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1GitRepoVolumeSource(BaseModel):
    """
    Represents a volume that is populated with the contents of a git repository. Git repo volumes do not support ownership management. Git repo volumes support SELinux relabeling.  DEPRECATED: GitRepo is deprecated. To provision a container with a git repo, mount an EmptyDir into an InitContainer that clones the repo using git, then mount the EmptyDir into the Pod's container.
    """ # noqa: E501
    directory: Optional[StrictStr] = Field(default=None, description="directory is the target directory name. Must not contain or start with '..'.  If '.' is supplied, the volume directory will be the git repository.  Otherwise, if specified, the volume will contain the git repository in the subdirectory with the given name.")
    repository: StrictStr = Field(description="repository is the URL")
    revision: Optional[StrictStr] = Field(default=None, description="revision is the commit hash for the specified revision.")
    __properties: ClassVar[List[str]] = ["directory", "repository", "revision"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1GitRepoVolumeSource from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1GitRepoVolumeSource from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "directory": obj.get("directory"),
            "repository": obj.get("repository") if obj.get("repository") is not None else '',
            "revision": obj.get("revision")
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApimachineryPkgApisMetaV1APIResource(BaseModel):
    """
    APIResource specifies the name of a resource and whether it is namespaced.
    """ # noqa: E501
    categories: Optional[List[StrictStr]] = Field(default=None, description="categories is a list of the grouped resources this resource belongs to (e.g. 'all')")
    group: Optional[StrictStr] = Field(default=None, description="group is the preferred group of the resource.  Empty implies the group of the containing resource list. For subresources, this may have a different value, for example: Scale\".")
    kind: StrictStr = Field(description="kind is the kind for the resource (e.g. 'Foo' is the kind for a resource 'foo')")
    name: StrictStr = Field(description="name is the plural name of the resource.")
    namespaced: StrictBool = Field(description="namespaced indicates if a resource is namespaced or not.")
    short_names: Optional[List[StrictStr]] = Field(default=None, description="shortNames is a list of suggested short names of the resource.", alias="shortNames")
    singular_name: StrictStr = Field(description="singularName is the singular name of the resource.  This allows clients to handle plural and singular opaquely. The singularName is more correct for reporting status on a single item and both singular and plural are allowed from the kubectl CLI interface.", alias="singularName")
    storage_version_hash: Optional[StrictStr] = Field(default=None, description="The hash value of the storage version, the version this resource is converted to when written to the data store. Value must be treated as opaque by clients. Only equality comparison on the value is valid. This is an alpha feature and may change or be removed in the future. The field is populated by the apiserver only if the StorageVersionHash feature gate is enabled. This field will remain optional even if it graduates.", alias="storageVersionHash")
    verbs: List[StrictStr] = Field(description="verbs is a list of supported kube verbs (this includes get, list, watch, create, update, patch, delete, deletecollection, and proxy)")
    version: Optional[StrictStr] = Field(default=None, description="version is the preferred version of the resource.  Empty implies the version of the containing resource list For subresources, this may have a different value, for example: v1 (while inside a v1beta1 version of the core resource's group)\".")
    __properties: ClassVar[List[str]] = ["categories", "group", "kind", "name", "namespaced", "shortNames", "singularName", "storageVersionHash", "verbs", "version"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1APIResource from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1APIResource from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "categories": obj.get("categories"),
            "group": obj.get("group"),
            "kind": obj.get("kind") if obj.get("kind") is not None else '',
            "name": obj.get("name") if obj.get("name") is not None else '',
            "namespaced": obj.get("namespaced") if obj.get("namespaced") is not None else False,
            "shortNames": obj.get("shortNames"),
            "singularName": obj.get("singularName") if obj.get("singularName") is not None else '',
            "storageVersionHash": obj.get("storageVersionHash"),
            "verbs": obj.get("verbs"),
            "version": obj.get("version")
        })
        return _obj





from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DeleteTensorsRequest(_message.Message):
    __slots__ = ("tensor_ids",)
    TENSOR_IDS_FIELD_NUMBER: _ClassVar[int]
    tensor_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, tensor_ids: _Optional[_Iterable[int]] = ...) -> None: ...

class TensorMetadata(_message.Message):
    __slots__ = ("tensor_id", "shape", "dtype", "nbytes", "device_type", "stride", "storage_offset", "device_index", "tensor_ref")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    NBYTES_FIELD_NUMBER: _ClassVar[int]
    DEVICE_TYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    DEVICE_INDEX_FIELD_NUMBER: _ClassVar[int]
    TENSOR_REF_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    nbytes: int
    device_type: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    device_index: int
    tensor_ref: int
    def __init__(self, tensor_id: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., nbytes: _Optional[int] = ..., device_type: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., device_index: _Optional[int] = ..., tensor_ref: _Optional[int] = ...) -> None: ...

class TensorChunk(_message.Message):
    __slots__ = ("tensor_id", "chunk_number", "data", "total_chunks", "shape", "stride", "storage_offset", "dtype", "total_bytes", "metadata")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    CHUNK_NUMBER_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    TOTAL_CHUNKS_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    TOTAL_BYTES_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    chunk_number: int
    data: bytes
    total_chunks: int
    shape: _containers.RepeatedScalarFieldContainer[int]
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    dtype: str
    total_bytes: int
    metadata: TensorMetadata
    def __init__(self, tensor_id: _Optional[int] = ..., chunk_number: _Optional[int] = ..., data: _Optional[bytes] = ..., total_chunks: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., dtype: _Optional[str] = ..., total_bytes: _Optional[int] = ..., metadata: _Optional[_Union[TensorMetadata, _Mapping]] = ...) -> None: ...

class TensorResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class GetScalarRequest(_message.Message):
    __slots__ = ("tensor_id", "tensor_metadata")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    TENSOR_METADATA_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    tensor_metadata: _containers.RepeatedCompositeFieldContainer[TensorMetadata]
    def __init__(self, tensor_id: _Optional[int] = ..., tensor_metadata: _Optional[_Iterable[_Union[TensorMetadata, _Mapping]]] = ...) -> None: ...

class GetScalarResponse(_message.Message):
    __slots__ = ("float_value", "int_value", "bool_value")
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    float_value: float
    int_value: int
    bool_value: bool
    def __init__(self, float_value: _Optional[float] = ..., int_value: _Optional[int] = ..., bool_value: bool = ...) -> None: ...

class CreateTensorRequest(_message.Message):
    __slots__ = ("tensor_id", "shape", "dtype", "nbytes", "device_type", "stride", "storage_offset", "device_index", "tensor_ref")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    NBYTES_FIELD_NUMBER: _ClassVar[int]
    DEVICE_TYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    DEVICE_INDEX_FIELD_NUMBER: _ClassVar[int]
    TENSOR_REF_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    nbytes: int
    device_type: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    device_index: int
    tensor_ref: int
    def __init__(self, tensor_id: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., nbytes: _Optional[int] = ..., device_type: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., device_index: _Optional[int] = ..., tensor_ref: _Optional[int] = ...) -> None: ...

class GetTensorRequest(_message.Message):
    __slots__ = ("tensor_id", "shape", "dtype", "stride", "storage_offset", "metadata")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    metadata: TensorMetadata
    def __init__(self, tensor_id: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., metadata: _Optional[_Union[TensorMetadata, _Mapping]] = ...) -> None: ...

class CopyTensorRequest(_message.Message):
    __slots__ = ("src_tensor_id", "dst_tensor_id", "src_offset", "dst_offset", "num_bytes", "src_metadata", "dst_metadata")
    SRC_TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    DST_TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    SRC_OFFSET_FIELD_NUMBER: _ClassVar[int]
    DST_OFFSET_FIELD_NUMBER: _ClassVar[int]
    NUM_BYTES_FIELD_NUMBER: _ClassVar[int]
    SRC_METADATA_FIELD_NUMBER: _ClassVar[int]
    DST_METADATA_FIELD_NUMBER: _ClassVar[int]
    src_tensor_id: int
    dst_tensor_id: int
    src_offset: int
    dst_offset: int
    num_bytes: int
    src_metadata: TensorMetadata
    dst_metadata: TensorMetadata
    def __init__(self, src_tensor_id: _Optional[int] = ..., dst_tensor_id: _Optional[int] = ..., src_offset: _Optional[int] = ..., dst_offset: _Optional[int] = ..., num_bytes: _Optional[int] = ..., src_metadata: _Optional[_Union[TensorMetadata, _Mapping]] = ..., dst_metadata: _Optional[_Union[TensorMetadata, _Mapping]] = ...) -> None: ...

class TensorReference(_message.Message):
    __slots__ = ("tensor_id",)
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    def __init__(self, tensor_id: _Optional[int] = ...) -> None: ...

class AtenArgument(_message.Message):
    __slots__ = ("tensor", "scalar_float", "scalar_int", "scalar_bool", "scalar_string", "list_value", "none_value", "scalar_dtype", "scalar_memory_format", "scalar_layout")
    TENSOR_FIELD_NUMBER: _ClassVar[int]
    SCALAR_FLOAT_FIELD_NUMBER: _ClassVar[int]
    SCALAR_INT_FIELD_NUMBER: _ClassVar[int]
    SCALAR_BOOL_FIELD_NUMBER: _ClassVar[int]
    SCALAR_STRING_FIELD_NUMBER: _ClassVar[int]
    LIST_VALUE_FIELD_NUMBER: _ClassVar[int]
    NONE_VALUE_FIELD_NUMBER: _ClassVar[int]
    SCALAR_DTYPE_FIELD_NUMBER: _ClassVar[int]
    SCALAR_MEMORY_FORMAT_FIELD_NUMBER: _ClassVar[int]
    SCALAR_LAYOUT_FIELD_NUMBER: _ClassVar[int]
    tensor: TensorReference
    scalar_float: float
    scalar_int: int
    scalar_bool: bool
    scalar_string: str
    list_value: AtenArgumentList
    none_value: bool
    scalar_dtype: str
    scalar_memory_format: str
    scalar_layout: str
    def __init__(self, tensor: _Optional[_Union[TensorReference, _Mapping]] = ..., scalar_float: _Optional[float] = ..., scalar_int: _Optional[int] = ..., scalar_bool: bool = ..., scalar_string: _Optional[str] = ..., list_value: _Optional[_Union[AtenArgumentList, _Mapping]] = ..., none_value: bool = ..., scalar_dtype: _Optional[str] = ..., scalar_memory_format: _Optional[str] = ..., scalar_layout: _Optional[str] = ...) -> None: ...

class AtenArgumentList(_message.Message):
    __slots__ = ("values", "is_tuple")
    VALUES_FIELD_NUMBER: _ClassVar[int]
    IS_TUPLE_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[AtenArgument]
    is_tuple: bool
    def __init__(self, values: _Optional[_Iterable[_Union[AtenArgument, _Mapping]]] = ..., is_tuple: bool = ...) -> None: ...

class ExecuteAtenRequest(_message.Message):
    __slots__ = ("op_name", "args", "outputs", "kwargs", "tensor_metadata", "output_metadata")
    class KwargsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: AtenArgument
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[AtenArgument, _Mapping]] = ...) -> None: ...
    OP_NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    TENSOR_METADATA_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_METADATA_FIELD_NUMBER: _ClassVar[int]
    op_name: str
    args: _containers.RepeatedCompositeFieldContainer[AtenArgument]
    outputs: _containers.RepeatedCompositeFieldContainer[TensorReference]
    kwargs: _containers.MessageMap[str, AtenArgument]
    tensor_metadata: _containers.RepeatedCompositeFieldContainer[TensorMetadata]
    output_metadata: _containers.RepeatedCompositeFieldContainer[TensorMetadata]
    def __init__(self, op_name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[AtenArgument, _Mapping]]] = ..., outputs: _Optional[_Iterable[_Union[TensorReference, _Mapping]]] = ..., kwargs: _Optional[_Mapping[str, AtenArgument]] = ..., tensor_metadata: _Optional[_Iterable[_Union[TensorMetadata, _Mapping]]] = ..., output_metadata: _Optional[_Iterable[_Union[TensorMetadata, _Mapping]]] = ...) -> None: ...

class ExecuteAtenResponse(_message.Message):
    __slots__ = ("success", "message", "output_tensors")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_TENSORS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    output_tensors: _containers.RepeatedCompositeFieldContainer[TensorReference]
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., output_tensors: _Optional[_Iterable[_Union[TensorReference, _Mapping]]] = ...) -> None: ...

class UpdateTensorRequest(_message.Message):
    __slots__ = ("tensor_id", "data", "shape", "dtype", "stride", "storage_offset", "metadata")
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    tensor_id: int
    data: bytes
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    metadata: TensorMetadata
    def __init__(self, tensor_id: _Optional[int] = ..., data: _Optional[bytes] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., metadata: _Optional[_Union[TensorMetadata, _Mapping]] = ...) -> None: ...

class GetTensorResponse(_message.Message):
    __slots__ = ("success", "message", "data", "shape", "dtype", "stride", "storage_offset")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    data: bytes
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., data: _Optional[bytes] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ...) -> None: ...

class ExecuteFunctionRequest(_message.Message):
    __slots__ = ("args", "kwargs", "callable_source", "callable_name", "retain_model")
    ARGS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    CALLABLE_SOURCE_FIELD_NUMBER: _ClassVar[int]
    CALLABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    RETAIN_MODEL_FIELD_NUMBER: _ClassVar[int]
    args: bytes
    kwargs: bytes
    callable_source: str
    callable_name: str
    retain_model: bool
    def __init__(self, args: _Optional[bytes] = ..., kwargs: _Optional[bytes] = ..., callable_source: _Optional[str] = ..., callable_name: _Optional[str] = ..., retain_model: bool = ...) -> None: ...

class RemoteTensorInfo(_message.Message):
    __slots__ = ("name", "storage_id", "shape", "dtype", "stride", "storage_offset", "storage_nbytes", "device_type", "device_index")
    NAME_FIELD_NUMBER: _ClassVar[int]
    STORAGE_ID_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_OFFSET_FIELD_NUMBER: _ClassVar[int]
    STORAGE_NBYTES_FIELD_NUMBER: _ClassVar[int]
    DEVICE_TYPE_FIELD_NUMBER: _ClassVar[int]
    DEVICE_INDEX_FIELD_NUMBER: _ClassVar[int]
    name: str
    storage_id: int
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: str
    stride: _containers.RepeatedScalarFieldContainer[int]
    storage_offset: int
    storage_nbytes: int
    device_type: str
    device_index: int
    def __init__(self, name: _Optional[str] = ..., storage_id: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ..., dtype: _Optional[str] = ..., stride: _Optional[_Iterable[int]] = ..., storage_offset: _Optional[int] = ..., storage_nbytes: _Optional[int] = ..., device_type: _Optional[str] = ..., device_index: _Optional[int] = ...) -> None: ...

class ExecuteFunctionResponse(_message.Message):
    __slots__ = ("success", "error_message", "tensors", "model_id")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    TENSORS_FIELD_NUMBER: _ClassVar[int]
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error_message: str
    tensors: _containers.RepeatedCompositeFieldContainer[RemoteTensorInfo]
    model_id: int
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ..., tensors: _Optional[_Iterable[_Union[RemoteTensorInfo, _Mapping]]] = ..., model_id: _Optional[int] = ...) -> None: ...

class ExecuteFunctionLog(_message.Message):
    __slots__ = ("stream", "text")
    STREAM_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    stream: str
    text: str
    def __init__(self, stream: _Optional[str] = ..., text: _Optional[str] = ...) -> None: ...

class ExecuteFunctionEvent(_message.Message):
    __slots__ = ("log", "result")
    LOG_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    log: ExecuteFunctionLog
    result: ExecuteFunctionResponse
    def __init__(self, log: _Optional[_Union[ExecuteFunctionLog, _Mapping]] = ..., result: _Optional[_Union[ExecuteFunctionResponse, _Mapping]] = ...) -> None: ...

class RegisterTensorsRequest(_message.Message):
    __slots__ = ("registrations",)
    REGISTRATIONS_FIELD_NUMBER: _ClassVar[int]
    registrations: _containers.RepeatedCompositeFieldContainer[TensorRegistration]
    def __init__(self, registrations: _Optional[_Iterable[_Union[TensorRegistration, _Mapping]]] = ...) -> None: ...

class TensorRegistration(_message.Message):
    __slots__ = ("storage_id", "tensor_id")
    STORAGE_ID_FIELD_NUMBER: _ClassVar[int]
    TENSOR_ID_FIELD_NUMBER: _ClassVar[int]
    storage_id: int
    tensor_id: int
    def __init__(self, storage_id: _Optional[int] = ..., tensor_id: _Optional[int] = ...) -> None: ...

class ExecuteModuleForwardRequest(_message.Message):
    __slots__ = ("model_id", "module_path", "input_tensor_ids", "input_metadata", "output_tensor_ids", "output_metadata")
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    MODULE_PATH_FIELD_NUMBER: _ClassVar[int]
    INPUT_TENSOR_IDS_FIELD_NUMBER: _ClassVar[int]
    INPUT_METADATA_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_TENSOR_IDS_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_METADATA_FIELD_NUMBER: _ClassVar[int]
    model_id: int
    module_path: str
    input_tensor_ids: _containers.RepeatedScalarFieldContainer[int]
    input_metadata: _containers.RepeatedCompositeFieldContainer[TensorMetadata]
    output_tensor_ids: _containers.RepeatedScalarFieldContainer[int]
    output_metadata: _containers.RepeatedCompositeFieldContainer[TensorMetadata]
    def __init__(self, model_id: _Optional[int] = ..., module_path: _Optional[str] = ..., input_tensor_ids: _Optional[_Iterable[int]] = ..., input_metadata: _Optional[_Iterable[_Union[TensorMetadata, _Mapping]]] = ..., output_tensor_ids: _Optional[_Iterable[int]] = ..., output_metadata: _Optional[_Iterable[_Union[TensorMetadata, _Mapping]]] = ...) -> None: ...

class ExecuteModuleForwardResponse(_message.Message):
    __slots__ = ("success", "error_message", "output_tensors")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_TENSORS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error_message: str
    output_tensors: _containers.RepeatedCompositeFieldContainer[RemoteTensorInfo]
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ..., output_tensors: _Optional[_Iterable[_Union[RemoteTensorInfo, _Mapping]]] = ...) -> None: ...

class ReleaseModelRequest(_message.Message):
    __slots__ = ("model_id",)
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    model_id: int
    def __init__(self, model_id: _Optional[int] = ...) -> None: ...

class BatchedExecuteAtenRequest(_message.Message):
    __slots__ = ("operations",)
    OPERATIONS_FIELD_NUMBER: _ClassVar[int]
    operations: _containers.RepeatedCompositeFieldContainer[ExecuteAtenRequest]
    def __init__(self, operations: _Optional[_Iterable[_Union[ExecuteAtenRequest, _Mapping]]] = ...) -> None: ...

class StreamRequest(_message.Message):
    __slots__ = ("execute_aten", "delete_tensors", "copy_tensor", "update_tensor", "get_tensor", "register_tensors", "batched_execute_aten", "raw_execute_aten", "raw_batched_execute_aten", "get_scalar", "execute_module_forward", "release_model", "chunk_number", "total_chunks", "total_bytes")
    EXECUTE_ATEN_FIELD_NUMBER: _ClassVar[int]
    DELETE_TENSORS_FIELD_NUMBER: _ClassVar[int]
    COPY_TENSOR_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TENSOR_FIELD_NUMBER: _ClassVar[int]
    GET_TENSOR_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TENSORS_FIELD_NUMBER: _ClassVar[int]
    BATCHED_EXECUTE_ATEN_FIELD_NUMBER: _ClassVar[int]
    RAW_EXECUTE_ATEN_FIELD_NUMBER: _ClassVar[int]
    RAW_BATCHED_EXECUTE_ATEN_FIELD_NUMBER: _ClassVar[int]
    GET_SCALAR_FIELD_NUMBER: _ClassVar[int]
    EXECUTE_MODULE_FORWARD_FIELD_NUMBER: _ClassVar[int]
    RELEASE_MODEL_FIELD_NUMBER: _ClassVar[int]
    CHUNK_NUMBER_FIELD_NUMBER: _ClassVar[int]
    TOTAL_CHUNKS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_BYTES_FIELD_NUMBER: _ClassVar[int]
    execute_aten: ExecuteAtenRequest
    delete_tensors: DeleteTensorsRequest
    copy_tensor: CopyTensorRequest
    update_tensor: UpdateTensorRequest
    get_tensor: GetTensorRequest
    register_tensors: RegisterTensorsRequest
    batched_execute_aten: BatchedExecuteAtenRequest
    raw_execute_aten: bytes
    raw_batched_execute_aten: bytes
    get_scalar: GetScalarRequest
    execute_module_forward: ExecuteModuleForwardRequest
    release_model: ReleaseModelRequest
    chunk_number: int
    total_chunks: int
    total_bytes: int
    def __init__(self, execute_aten: _Optional[_Union[ExecuteAtenRequest, _Mapping]] = ..., delete_tensors: _Optional[_Union[DeleteTensorsRequest, _Mapping]] = ..., copy_tensor: _Optional[_Union[CopyTensorRequest, _Mapping]] = ..., update_tensor: _Optional[_Union[UpdateTensorRequest, _Mapping]] = ..., get_tensor: _Optional[_Union[GetTensorRequest, _Mapping]] = ..., register_tensors: _Optional[_Union[RegisterTensorsRequest, _Mapping]] = ..., batched_execute_aten: _Optional[_Union[BatchedExecuteAtenRequest, _Mapping]] = ..., raw_execute_aten: _Optional[bytes] = ..., raw_batched_execute_aten: _Optional[bytes] = ..., get_scalar: _Optional[_Union[GetScalarRequest, _Mapping]] = ..., execute_module_forward: _Optional[_Union[ExecuteModuleForwardRequest, _Mapping]] = ..., release_model: _Optional[_Union[ReleaseModelRequest, _Mapping]] = ..., chunk_number: _Optional[int] = ..., total_chunks: _Optional[int] = ..., total_bytes: _Optional[int] = ...) -> None: ...

class StreamResponse(_message.Message):
    __slots__ = ("success", "error_message", "execute_aten", "delete_tensors", "copy_tensor", "update_tensor", "get_tensor", "register_tensors", "get_scalar", "execute_module_forward", "chunk_number", "total_chunks", "server_backlog_ns", "server_handle_ns")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    EXECUTE_ATEN_FIELD_NUMBER: _ClassVar[int]
    DELETE_TENSORS_FIELD_NUMBER: _ClassVar[int]
    COPY_TENSOR_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TENSOR_FIELD_NUMBER: _ClassVar[int]
    GET_TENSOR_FIELD_NUMBER: _ClassVar[int]
    REGISTER_TENSORS_FIELD_NUMBER: _ClassVar[int]
    GET_SCALAR_FIELD_NUMBER: _ClassVar[int]
    EXECUTE_MODULE_FORWARD_FIELD_NUMBER: _ClassVar[int]
    CHUNK_NUMBER_FIELD_NUMBER: _ClassVar[int]
    TOTAL_CHUNKS_FIELD_NUMBER: _ClassVar[int]
    SERVER_BACKLOG_NS_FIELD_NUMBER: _ClassVar[int]
    SERVER_HANDLE_NS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error_message: str
    execute_aten: ExecuteAtenResponse
    delete_tensors: TensorResponse
    copy_tensor: TensorResponse
    update_tensor: TensorResponse
    get_tensor: GetTensorResponse
    register_tensors: TensorResponse
    get_scalar: GetScalarResponse
    execute_module_forward: ExecuteModuleForwardResponse
    chunk_number: int
    total_chunks: int
    server_backlog_ns: int
    server_handle_ns: int
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ..., execute_aten: _Optional[_Union[ExecuteAtenResponse, _Mapping]] = ..., delete_tensors: _Optional[_Union[TensorResponse, _Mapping]] = ..., copy_tensor: _Optional[_Union[TensorResponse, _Mapping]] = ..., update_tensor: _Optional[_Union[TensorResponse, _Mapping]] = ..., get_tensor: _Optional[_Union[GetTensorResponse, _Mapping]] = ..., register_tensors: _Optional[_Union[TensorResponse, _Mapping]] = ..., get_scalar: _Optional[_Union[GetScalarResponse, _Mapping]] = ..., execute_module_forward: _Optional[_Union[ExecuteModuleForwardResponse, _Mapping]] = ..., chunk_number: _Optional[int] = ..., total_chunks: _Optional[int] = ..., server_backlog_ns: _Optional[int] = ..., server_handle_ns: _Optional[int] = ...) -> None: ...



"""
NVIDIA GPU metrics source using NVIDIA Management Library (NVML).
"""
import logging
from typing import Dict, List, Optional

from skytorch.server.metrics.source import MetricsSource, MetricData

logger = logging.getLogger(__name__)


class NvidiaGPUMetricsSource(MetricsSource):
    """
    Metrics source for NVIDIA GPUs using pynvml (Python bindings to NVML).

    Collects GPU metrics such as:
    - GPU utilization (compute and memory)
    - Temperature
    - Power usage
    - Memory usage
    - Clock speeds
    - Fan speed
    - Performance state
    """

    def __init__(self):
        """Initialize the NVIDIA GPU metrics source."""
        self._initialized = False
        self._device_count = 0
        self._nvml = None

        try:
            import pynvml
            self._nvml = pynvml
            self._nvml.nvmlInit()
            self._device_count = self._nvml.nvmlDeviceGetCount()
            self._initialized = True
            logger.info(f"NVIDIA GPU metrics source initialized with {self._device_count} device(s)")
        except ImportError:
            logger.warning(
                "pynvml not available. Install with: pip install nvidia-ml-py"
            )
        except Exception as e:
            logger.warning(f"Failed to initialize NVML: {e}")

    def get_source_name(self) -> str:
        """Get the source name."""
        return "nvidia-gpu"

    def is_available(self) -> bool:
        """Check if NVIDIA GPUs are available."""
        return self._initialized and self._device_count > 0

    def collect(
        self,
        *,
        metric_names: Optional[List[str]] = None,
        label_filters: Optional[Dict[str, str]] = None
    ) -> List[MetricData]:
        """
        Collect NVIDIA GPU metrics.

        Args:
            metric_names: Optional list of specific metrics to collect
            label_filters: Optional label filters

        Returns:
            List of metric data points
        """
        if not self.is_available():
            return []

        metrics = []

        # Collect metrics for each GPU
        for i in range(self._device_count):
            try:
                handle = self._nvml.nvmlDeviceGetHandleByIndex(i)
                device_name = self._nvml.nvmlDeviceGetName(handle)

                # Base labels for all metrics from this GPU
                base_labels = {
                    "device": str(i),
                    "gpu_name": device_name,
                }

                # Apply label filters if provided
                if label_filters:
                    if not all(base_labels.get(k) == v for k, v in label_filters.items()):
                        continue

                # Collect GPU utilization
                if self._should_collect("gpu.utilization.compute", metric_names):
                    try:
                        utilization = self._nvml.nvmlDeviceGetUtilizationRates(handle)
                        metrics.append(MetricData(
                            name="gpu.utilization.compute",
                            value=float(utilization.gpu),
                            metric_type="GAUGE",
                            unit="percent",
                            labels=base_labels,
                            help_text="GPU compute utilization"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get GPU utilization for device {i}: {e}")

                # Collect memory utilization
                if self._should_collect("gpu.utilization.memory", metric_names):
                    try:
                        utilization = self._nvml.nvmlDeviceGetUtilizationRates(handle)
                        metrics.append(MetricData(
                            name="gpu.utilization.memory",
                            value=float(utilization.memory),
                            metric_type="GAUGE",
                            unit="percent",
                            labels=base_labels,
                            help_text="GPU memory bandwidth utilization"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get memory utilization for device {i}: {e}")

                # Collect temperature
                if self._should_collect("gpu.temperature", metric_names):
                    try:
                        temp = self._nvml.nvmlDeviceGetTemperature(
                            handle,
                            self._nvml.NVML_TEMPERATURE_GPU
                        )
                        metrics.append(MetricData(
                            name="gpu.temperature",
                            value=float(temp),
                            metric_type="GAUGE",
                            unit="celsius",
                            labels=base_labels,
                            help_text="GPU temperature"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get temperature for device {i}: {e}")

                # Collect power usage
                if self._should_collect("gpu.power.usage", metric_names):
                    try:
                        power = self._nvml.nvmlDeviceGetPowerUsage(handle)
                        # Convert milliwatts to watts
                        metrics.append(MetricData(
                            name="gpu.power.usage",
                            value=float(power) / 1000.0,
                            metric_type="GAUGE",
                            unit="watts",
                            labels=base_labels,
                            help_text="GPU power usage"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get power usage for device {i}: {e}")

                # Collect power limit
                if self._should_collect("gpu.power.limit", metric_names):
                    try:
                        power_limit = self._nvml.nvmlDeviceGetPowerManagementLimit(handle)
                        # Convert milliwatts to watts
                        metrics.append(MetricData(
                            name="gpu.power.limit",
                            value=float(power_limit) / 1000.0,
                            metric_type="GAUGE",
                            unit="watts",
                            labels=base_labels,
                            help_text="GPU power limit"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get power limit for device {i}: {e}")

                # Collect memory info
                if self._should_collect("gpu.memory.used", metric_names) or \
                   self._should_collect("gpu.memory.free", metric_names) or \
                   self._should_collect("gpu.memory.total", metric_names):
                    try:
                        mem_info = self._nvml.nvmlDeviceGetMemoryInfo(handle)

                        if self._should_collect("gpu.memory.used", metric_names):
                            metrics.append(MetricData(
                                name="gpu.memory.used",
                                value=float(mem_info.used),
                                metric_type="GAUGE",
                                unit="bytes",
                                labels=base_labels,
                                help_text="GPU memory used"
                            ))

                        if self._should_collect("gpu.memory.free", metric_names):
                            metrics.append(MetricData(
                                name="gpu.memory.free",
                                value=float(mem_info.free),
                                metric_type="GAUGE",
                                unit="bytes",
                                labels=base_labels,
                                help_text="GPU memory free"
                            ))

                        if self._should_collect("gpu.memory.total", metric_names):
                            metrics.append(MetricData(
                                name="gpu.memory.total",
                                value=float(mem_info.total),
                                metric_type="GAUGE",
                                unit="bytes",
                                labels=base_labels,
                                help_text="GPU total memory"
                            ))
                    except Exception as e:
                        logger.debug(f"Failed to get memory info for device {i}: {e}")

                # Collect clock speeds
                if self._should_collect("gpu.clock.graphics", metric_names):
                    try:
                        clock = self._nvml.nvmlDeviceGetClockInfo(
                            handle,
                            self._nvml.NVML_CLOCK_GRAPHICS
                        )
                        metrics.append(MetricData(
                            name="gpu.clock.graphics",
                            value=float(clock),
                            metric_type="GAUGE",
                            unit="mhz",
                            labels=base_labels,
                            help_text="GPU graphics clock speed"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get graphics clock for device {i}: {e}")

                if self._should_collect("gpu.clock.sm", metric_names):
                    try:
                        clock = self._nvml.nvmlDeviceGetClockInfo(
                            handle,
                            self._nvml.NVML_CLOCK_SM
                        )
                        metrics.append(MetricData(
                            name="gpu.clock.sm",
                            value=float(clock),
                            metric_type="GAUGE",
                            unit="mhz",
                            labels=base_labels,
                            help_text="GPU SM (streaming multiprocessor) clock speed"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get SM clock for device {i}: {e}")

                if self._should_collect("gpu.clock.memory", metric_names):
                    try:
                        clock = self._nvml.nvmlDeviceGetClockInfo(
                            handle,
                            self._nvml.NVML_CLOCK_MEM
                        )
                        metrics.append(MetricData(
                            name="gpu.clock.memory",
                            value=float(clock),
                            metric_type="GAUGE",
                            unit="mhz",
                            labels=base_labels,
                            help_text="GPU memory clock speed"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get memory clock for device {i}: {e}")

                # Collect fan speed
                if self._should_collect("gpu.fan.speed", metric_names):
                    try:
                        fan_speed = self._nvml.nvmlDeviceGetFanSpeed(handle)
                        metrics.append(MetricData(
                            name="gpu.fan.speed",
                            value=float(fan_speed),
                            metric_type="GAUGE",
                            unit="percent",
                            labels=base_labels,
                            help_text="GPU fan speed"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get fan speed for device {i}: {e}")

                # Collect performance state
                if self._should_collect("gpu.performance.state", metric_names):
                    try:
                        pstate = self._nvml.nvmlDeviceGetPerformanceState(handle)
                        metrics.append(MetricData(
                            name="gpu.performance.state",
                            value=float(pstate),
                            metric_type="GAUGE",
                            unit="",
                            labels=base_labels,
                            help_text="GPU performance state (0=max, 15=min)"
                        ))
                    except Exception as e:
                        logger.debug(f"Failed to get performance state for device {i}: {e}")

            except Exception as e:
                logger.error(f"Error collecting metrics for GPU {i}: {e}")

        return metrics

    def _should_collect(self, metric_name: str, filter_names: Optional[List[str]]) -> bool:
        """
        Check if a metric should be collected based on filter.

        Args:
            metric_name: Name of the metric
            filter_names: Optional list of metric names to collect

        Returns:
            True if the metric should be collected
        """
        if filter_names is None:
            return True
        return metric_name in filter_names

    def get_metadata(self) -> Dict[str, str]:
        """Get metadata about NVIDIA GPUs."""
        if not self.is_available():
            return {}

        metadata = {
            "device_count": str(self._device_count),
        }

        try:
            metadata["driver_version"] = self._nvml.nvmlSystemGetDriverVersion()
            metadata["nvml_version"] = self._nvml.nvmlSystemGetNVMLVersion()
        except Exception as e:
            logger.debug(f"Failed to get NVML metadata: {e}")

        return metadata

    def cleanup(self) -> None:
        """Cleanup NVML resources."""
        if self._initialized:
            try:
                self._nvml.nvmlShutdown()
                logger.info("NVIDIA GPU metrics source cleaned up")
            except Exception as e:
                logger.error(f"Error during NVML cleanup: {e}")



"""Utilities for extracting and preparing callable source code for remote execution."""

import inspect
import textwrap
import types


def generate_imports(fn) -> str:
    """Generate import statements for global references used by a function."""
    lines = []
    seen = set()
    for name in fn.__code__.co_names:
        if name in seen or name not in fn.__globals__:
            continue
        seen.add(name)
        obj = fn.__globals__[name]
        if isinstance(obj, types.ModuleType):
            if obj.__name__ != name:
                lines.append(f"import {obj.__name__} as {name}")
            else:
                lines.append(f"import {obj.__name__}")
        elif hasattr(obj, "__module__") and hasattr(obj, "__qualname__"):
            module = getattr(obj, "__module__", None)
            if module and module != "builtins":
                lines.append(f"from {module} import {name}")
    return "\n".join(lines)


def get_callable_source(fn) -> tuple[str, str]:
    """Extract source code and name from a callable.

    Returns:
        A tuple of (source_code, function_name).

    Raises:
        TypeError: If the callable cannot be serialized for remote execution.
    """
    if getattr(fn, "__name__", "") == "<lambda>":
        raise TypeError(
            "Lambdas cannot be serialized for remote execution. " "Use a named function instead."
        )

    if fn.__code__.co_freevars:
        raise TypeError(
            "Closures cannot be serialized for remote execution. "
            "Move captured variables to function arguments."
        )

    try:
        source = inspect.getsource(fn)
        source = textwrap.dedent(source)
    except (OSError, TypeError) as e:
        raise TypeError(
            f"Cannot extract source code for {fn!r}. Functions defined in "
            "interactive sessions or C extensions are not supported."
        ) from e

    imports = generate_imports(fn)
    if imports:
        source = imports + "\n\n" + source

    return source, fn.__name__



"""
SkyTorch ATen Dropout Operations.

Handle deterministic native_dropout edge cases client-side to avoid
server round-trips that may silently fail in streaming mode:

- p=0 or train=False: returns input unchanged with all-True mask
- p=1 with train=True: returns all-zeros with all-False mask

The general case (0 < p < 1) still delegates to the server via the
standard fallback.
"""

from typing import Optional

import torch

from .dispatch import _sky_kernel_fallback


def _native_dropout(
    input: torch.Tensor, p: float, train: Optional[bool]
) -> tuple[torch.Tensor, torch.Tensor]:
    if train is False or p == 0:
        return (input, torch.ones_like(input, dtype=torch.bool))
    if p == 1:
        return (torch.zeros_like(input), torch.zeros_like(input, dtype=torch.bool))
    return _sky_kernel_fallback(
        torch.ops.aten.native_dropout.default, input, p, train
    )



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApimachineryPkgApisMetaV1ManagedFieldsEntry(BaseModel):
    """
    ManagedFieldsEntry is a workflow-id, a FieldSet and the group version of the resource that the fieldset applies to.
    """ # noqa: E501
    api_version: Optional[StrictStr] = Field(default=None, description="APIVersion defines the version of this resource that this field set applies to. The format is \"group/version\" just like the top-level APIVersion field. It is necessary to track the version of a field set because it cannot be automatically converted.", alias="apiVersion")
    fields_type: Optional[StrictStr] = Field(default=None, description="FieldsType is the discriminator for the different fields format and version. There is currently only one possible value: \"FieldsV1\"", alias="fieldsType")
    fields_v1: Optional[Dict[str, Any]] = Field(default=None, description="FieldsV1 holds the first JSON version format as described in the \"FieldsV1\" type.", alias="fieldsV1")
    manager: Optional[StrictStr] = Field(default=None, description="Manager is an identifier of the workflow managing these fields.")
    operation: Optional[StrictStr] = Field(default=None, description="Operation is the type of operation which lead to this ManagedFieldsEntry being created. The only valid values for this field are 'Apply' and 'Update'.")
    subresource: Optional[StrictStr] = Field(default=None, description="Subresource is the name of the subresource used to update that object, or empty string if the object was updated through the main resource. The value of this field is used to distinguish between managers, even if they share the same name. For example, a status update will be distinct from a regular update using the same manager name. Note that the APIVersion field is not related to the Subresource field and it always corresponds to the version of the main resource.")
    time: Optional[datetime] = Field(default=None, description="Time is the timestamp of when the ManagedFields entry was added. The timestamp will also be updated if a field is added, the manager changes any of the owned fields value or removes a field. The timestamp does not update when a field is removed from the entry because another manager took it over.")
    __properties: ClassVar[List[str]] = ["apiVersion", "fieldsType", "fieldsV1", "manager", "operation", "subresource", "time"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1ManagedFieldsEntry from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1ManagedFieldsEntry from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "apiVersion": obj.get("apiVersion"),
            "fieldsType": obj.get("fieldsType"),
            "fieldsV1": obj.get("fieldsV1"),
            "manager": obj.get("manager"),
            "operation": obj.get("operation"),
            "subresource": obj.get("subresource"),
            "time": obj.get("time")
        })
        return _obj





"""
StreamManager for bidirectional gRPC streaming.

This module provides thread-safe bidirectional stream management for
low-latency tensor operations, enabling pipelining of requests without
waiting for individual responses.

All operations (execute_aten, copy_tensor, update_tensor, get_tensor, etc.)
go through a single ordered stream to ensure proper sequencing.
"""

from __future__ import annotations

import asyncio
import collections
import gc
import logging
import os
import struct
import sys
import threading
import time
from dataclasses import dataclass
from typing import Optional

from skytorch.torch.profiler import PROFILING_ENABLED

try:
    import grpc
except ImportError as e:
    raise ImportError(f"grpcio package is required: {e}")

try:
    from skytorch.torch.server import service_pb2
    from skytorch.torch.server import service_pb2_grpc
except ImportError:
    raise ImportError("Generated gRPC code not found. Run hack/gen-grpc-proto.sh first.")

from grpc.aio._typing import MetadataType

logger = logging.getLogger(__name__)

# Chunk size for large tensor payloads (1MB, matching serialization module)
STREAM_CHUNK_SIZE = 1024 * 1024


def _serialize_tensor_metadata_binary(meta) -> bytes:
    """Serialize a TensorMetadata proto to the ATen binary tensor metadata format.

    Format: [tensor_id:u64][ndim:u8][shape:i64×ndim][stride:i64×ndim]
            [dtype_len:u8][dtype:bytes][storage_offset:i64][nbytes:i64]
            [device_type_len:u8][device_type:bytes][device_index:i32]
            [has_tensor_ref:u8][tensor_ref:u64 if has]
    """
    shape = list(meta.shape)
    stride = list(meta.stride) if meta.stride else [0] * len(shape)
    ndim = len(shape)
    dtype_bytes = meta.dtype.encode("utf-8")
    device_type_bytes = meta.device_type.encode("utf-8")
    has_ref = meta.HasField("tensor_ref")

    parts = [
        struct.pack(f"<QB{ndim}q{ndim}q", meta.tensor_id, ndim, *shape, *stride),
        struct.pack(f"<B{len(dtype_bytes)}s", len(dtype_bytes), dtype_bytes),
        struct.pack("<qq", meta.storage_offset, meta.nbytes),
        struct.pack(
            f"<B{len(device_type_bytes)}si",
            len(device_type_bytes),
            device_type_bytes,
            meta.device_index,
        ),
    ]
    if has_ref:
        parts.append(struct.pack("<BQ", 1, meta.tensor_ref))
    else:
        parts.append(b"\x00")
    return b"".join(parts)


@dataclass
class PendingRequest:
    """Tracks a pending request awaiting response."""

    future: Optional[asyncio.Future]
    request_type: str


class StreamManager:
    """
    Bidirectional stream manager for pipelined operations.

    All tensor operations go through a single ordered stream to ensure
    proper sequencing. Responses are matched to requests by order (FIFO).

    Key features:
    - FIFO ordering of requests and responses
    - Efficient async queue (no busy-wait polling)
    - Automatic stream lifecycle management

    Note: All operations must be called from the asyncio event loop thread.
    """

    def __init__(
        self,
        stub: service_pb2_grpc.ServiceStub,
        metadata: Optional[MetadataType] = None,
    ):
        """
        Initialize the stream manager.

        Args:
            stub: gRPC service stub
            metadata: Optional metadata for requests
        """
        self._stub = stub
        self._metadata = metadata

        # Pending requests - FIFO queue of futures matched by order
        self._pending: collections.deque[PendingRequest] = collections.deque()

        # Stream state
        self._request_queue: Optional[asyncio.Queue] = None
        self._sender_task: Optional[asyncio.Task] = None
        self._receiver_task: Optional[asyncio.Task] = None
        self._stream_call: Optional[grpc.aio.StreamStreamCall] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._started = False
        self._closing = False

        # Deferred error from fire-and-forget operations
        self._last_error: Optional[Exception] = None

        # Batch buffer for execute_aten operations
        self._batch_buffer: list[service_pb2.ExecuteAtenRequest] = []
        self._flush_scheduled: bool = False
        self._batch_flush_timer = None

        # Batch buffer for raw binary execute_aten operations (C++ fast path)
        self._raw_batch_buffer: list[bytes] = []
        self._raw_flush_scheduled: bool = False
        self._raw_flush_timer = None

        # Main-thread submission buffer: collects ops from all submit methods
        # under a single lock, preserving FIFO ordering while reducing
        # call_soon_threadsafe calls by ~98%. Each entry is either:
        #   ("raw", bytes)    — raw binary execute_aten
        #   ("aten", ExecuteAtenRequest) — protobuf execute_aten
        #   ("req", StreamRequest) — non-batched request (copy, update, register)
        #   ("ff", list[int])      — deferred delete tensor IDs
        self._mt_ops: list[tuple] = []
        self._mt_lock = threading.RLock()
        self._mt_wake_pending: bool = False

        # Deferred delete tensor IDs: accumulated from fire-and-forget delete
        # requests that preserve FIFO ordering but don't force a batch flush.
        # They piggyback on the next batch flush instead of triggering one.
        # IDs are batched into a single DeleteTensorsRequest when flushed,
        # reducing per-message gRPC overhead.
        self._deferred_delete_ids: list[int] = []
        self._deferred_flush_scheduled: bool = False
        self._deferred_flush_timer = None

        # Shutdown signaling
        self._shutdown_event: Optional[asyncio.Event] = None

        # C++ raw submit buffer integration
        self._cpp_submit_available: bool = False

        # Free-threaded Python uses biased reference counting: decrefs on
        # non-owning threads are deferred and only reconciled during GC.
        # When deletes accumulate, we call gc.collect(0) to reconcile
        # biased refcounts — this may free additional objects whose
        # decrefs were deferred by non-owning threads (e.g., the
        # AsyncStreamer worker thread).  Triggered when live GPU storage
        # (allocated - freed) grows by the threshold since the last GC,
        # indicating biased-refcount tensors are accumulating.
        _is_ft_build = hasattr(sys, "_is_gil_enabled")
        self._gc_on_delete = _is_ft_build
        self._gc_live_bytes_at_last_gc: int = 0
        self._gc_idle_handle = None

    async def start(self, loop: asyncio.AbstractEventLoop) -> None:
        """
        Start the bidirectional stream.

        Args:
            loop: Event loop to run stream tasks on
        """
        if self._started:
            return

        self._loop = loop
        self._request_queue = asyncio.Queue()
        self._shutdown_event = asyncio.Event()

        # Start sender and receiver tasks
        self._sender_task = asyncio.create_task(self._sender_loop())
        self._receiver_task = asyncio.create_task(self._receiver_loop())
        self._started = True

    def _enqueue(self, pending, stream_request):
        """Enqueue a single request. Must run on the event loop thread."""
        if pending is not None:
            self._pending.append(pending)
        self._request_queue.put_nowait(stream_request)

    def _enqueue_batch(self, pending, stream_requests):
        """Enqueue multiple requests with one pending entry. Must run on the event loop thread."""
        for request in stream_requests:
            self._request_queue.put_nowait(request)
        if pending is not None:
            self._pending.append(pending)

    # Flush threshold: when this many ops are buffered, flush immediately.
    # Override with SKYTORCH_BATCH_THRESHOLD env var.
    _BATCH_FLUSH_THRESHOLD = int(os.environ.get("SKYTORCH_BATCH_THRESHOLD", "64"))

    # Coalescing delay: defer partial-batch flush by this many seconds,
    # allowing more ops to accumulate and amortize per-message gRPC overhead.
    # At C++ dispatch rate (~17 us/op), 1 ms ≈ 59 ops per batch.
    # Sync points force an immediate flush, so this never adds sync latency.
    # Override with SKYTORCH_BATCH_COALESCE_MS env var (in milliseconds).
    _BATCH_COALESCE_DELAY = float(os.environ.get("SKYTORCH_BATCH_COALESCE_MS", "2")) / 1000.0

    # Delete coalescing: defer delete flushes by this many seconds, allowing
    # deletes from multiple GC waves to accumulate into a single message.
    # Override with SKYTORCH_DELETE_COALESCE_MS env var (in milliseconds).
    _DELETE_COALESCE_DELAY = float(os.environ.get("SKYTORCH_DELETE_COALESCE_MS", "10")) / 1000.0

    # Safety threshold: flush deletes immediately when this many IDs accumulate,
    # preventing unbounded GPU memory retention during large GC events.
    # Override with SKYTORCH_DELETE_THRESHOLD env var.
    _DELETE_FLUSH_THRESHOLD = int(os.environ.get("SKYTORCH_DELETE_THRESHOLD", "256"))

    # GC reconciliation threshold: run gc.collect(0) when live GPU storage
    # (allocated - freed) grows by this many bytes since the last GC.
    # This directly measures biased-refcount tensor accumulation — tensors
    # allocated but not freed because deferred decrefs haven't been reconciled.
    # Default 1 GiB.  Only active on the free-threaded build.
    # Override with SKYTORCH_GC_DELETE_BYTES env var.
    _GC_DELETE_BYTES_THRESHOLD = int(os.environ.get("SKYTORCH_GC_DELETE_BYTES", str(1024 ** 3)))

    def _enqueue_execute_aten(self, request: service_pb2.ExecuteAtenRequest) -> None:
        """Buffer a protobuf execute_aten request for batching. Must run on the event loop thread."""
        self._batch_buffer.append(request)
        if len(self._batch_buffer) >= self._BATCH_FLUSH_THRESHOLD:
            self._flush_batch()
        elif not self._flush_scheduled:
            self._flush_scheduled = True
            self._batch_flush_timer = self._loop.call_later(
                self._BATCH_COALESCE_DELAY, self._flush_batch
            )

    def _enqueue_execute_aten_bytes(self, raw_bytes: bytes) -> None:
        """Buffer a raw binary execute_aten request for batching. Must run on the event loop thread."""
        self._raw_batch_buffer.append(raw_bytes)
        if len(self._raw_batch_buffer) >= self._BATCH_FLUSH_THRESHOLD:
            self._flush_raw_batch()
        elif not self._raw_flush_scheduled:
            self._raw_flush_scheduled = True
            self._raw_flush_timer = self._loop.call_later(
                self._BATCH_COALESCE_DELAY, self._flush_raw_batch
            )

    def _flush_batch(self) -> None:
        """Flush buffered execute_aten requests as a batch. Must run on the event loop thread."""
        self._flush_scheduled = False
        if self._batch_flush_timer is not None:
            self._batch_flush_timer.cancel()
            self._batch_flush_timer = None
        if not self._batch_buffer:
            return

        if len(self._batch_buffer) == 1:
            # Single request: send as regular execute_aten to avoid wrapping overhead
            stream_request = service_pb2.StreamRequest(execute_aten=self._batch_buffer[0])
        else:
            # Multiple requests: send as batch
            batch = service_pb2.BatchedExecuteAtenRequest(operations=self._batch_buffer)
            stream_request = service_pb2.StreamRequest(batched_execute_aten=batch)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Flushing batch of {len(self._batch_buffer)} execute_aten ops")

        self._batch_buffer = []
        self._request_queue.put_nowait(stream_request)

    def _flush_raw_batch(self, piggyback_deletes: bool = True) -> None:
        """Flush buffered raw binary execute_aten requests. Must run on the event loop thread."""
        self._raw_flush_scheduled = False
        if self._raw_flush_timer is not None:
            self._raw_flush_timer.cancel()
            self._raw_flush_timer = None
        if not self._raw_batch_buffer:
            return

        if PROFILING_ENABLED:
            from skytorch.torch.profiler import ClientProfiler

            _prof = ClientProfiler.get()
            _batch_size = len(self._raw_batch_buffer)
            _prof.batch_count.add_count()
            _prof.batch_size_total += _batch_size
            if _batch_size > _prof.batch_size_max:
                _prof.batch_size_max = _batch_size

        if len(self._raw_batch_buffer) == 1:
            # Single request: send as raw_execute_aten
            stream_request = service_pb2.StreamRequest(raw_execute_aten=self._raw_batch_buffer[0])
        else:
            # Multiple requests: concatenate into raw_batched_execute_aten
            # Each request is prefixed with its length as uint32 LE
            import struct

            parts = []
            for raw in self._raw_batch_buffer:
                parts.append(struct.pack("<I", len(raw)))
                parts.append(raw)
            stream_request = service_pb2.StreamRequest(raw_batched_execute_aten=b"".join(parts))
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    f"Flushing raw batch of {len(self._raw_batch_buffer)} execute_aten ops"
                )

        self._raw_batch_buffer = []
        self._request_queue.put_nowait(stream_request)

        # Piggyback deferred deletes: the batch (which may reference these
        # tensors) was just sent, so it's safe to send deletes now.  This
        # avoids a separate flush that would fragment future batches.
        # Suppressed during sync flushes (get_scalar/get_tensor) where
        # deletes must go AFTER the sync request to avoid races.
        if piggyback_deletes:
            self._send_deferred_deletes()

    async def _sender_loop(self) -> None:
        """Send requests from queue to the stream."""
        try:

            async def request_generator():
                while not self._closing:
                    request = await self._request_queue.get()
                    if request is None:
                        break
                    yield request
                    # Drain ready items without re-awaiting
                    while not self._request_queue.empty():
                        request = self._request_queue.get_nowait()
                        if request is None:
                            return
                        yield request

            # Start the bidirectional stream
            self._stream_call = self._stub.StreamOperations(
                request_generator(), metadata=self._metadata
            )

            # Wait for shutdown signal
            await self._shutdown_event.wait()

        except grpc.RpcError as e:
            logger.error(f"Stream sender error: {e}")
            self._fail_all_pending(str(e))
        except Exception as e:
            logger.error(f"Unexpected stream sender error: {e}")
            self._fail_all_pending(str(e))

    async def _receiver_loop(self) -> None:
        """Receive responses from the stream and resolve futures in order."""
        try:
            # Wait for stream to be established
            while self._stream_call is None and not self._closing:
                await asyncio.sleep(0.01)

            if self._stream_call is None:
                return

            async for response in self._stream_call:
                if not self._pending:
                    logger.warning("Received response but no pending requests")
                    continue
                pending = self._pending.popleft()

                # Check for deferred error from fire-and-forget operations
                if response.error_message and not response.success:
                    error = RuntimeError(f"Stream operation failed: {response.error_message}")
                    if pending.future is not None:
                        pending.future.set_exception(error)
                    self.record_error(error)
                elif pending.future is not None:
                    pending.future.set_result(response)

        except grpc.RpcError as e:
            if not self._closing:
                logger.error(f"Stream receiver error: {e}")
                self._fail_all_pending(str(e))
        except Exception as e:
            if not self._closing:
                logger.error(f"Unexpected stream receiver error: {e}")
                self._fail_all_pending(str(e))

    def record_error(self, error: Exception) -> None:
        """Record error for deferred raising. First-error-wins semantics."""
        if self._last_error is None:
            self._last_error = error

    def check_error(self) -> None:
        """Raise deferred error if any, then clear it."""
        error = self._last_error
        if error is not None:
            self._last_error = None
            raise error

    def _fail_all_pending(self, error_message: str) -> None:
        """Fail all pending requests with an error."""
        for pending in self._pending:
            if pending.future is not None and not pending.future.done():
                pending.future.set_exception(RuntimeError(f"Stream error: {error_message}"))
        self._pending.clear()

    def _drain_cpp_into_mt_ops(self) -> None:
        """Drain C++ raw submit buffer into _mt_ops. Caller must hold _mt_lock.

        Uses C++ flush_raw_batches to build batched payloads in C++,
        reducing event loop GIL-holding time from ~200-500us (Python
        drain + struct.pack + bytes.join) to ~5-10us (C++ concatenation).
        """
        if not self._cpp_submit_available:
            return
        try:
            from skytorch.torch.backend._C import _flush_raw_batches

            batches = _flush_raw_batches(self._BATCH_FLUSH_THRESHOLD)
            for batch_bytes, count in batches:
                self._mt_ops.append(("cpp_batch", (batch_bytes, count)))
        except (ImportError, AttributeError):
            pass

    def _drain_mt_ops(self) -> None:
        """Drain the main-thread ops buffer. Must run on the event loop thread."""
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            if not self._mt_ops:
                self._mt_wake_pending = False
                return
            batch = self._mt_ops
            self._mt_ops = []
            self._mt_wake_pending = False
        self._process_mt_ops(batch)

        # Schedule an idle-GC: if no new ops arrive within 100ms, run
        # gc.collect(0) to reconcile any remaining biased refcounts.
        # This handles the gap between the last token of a turn and the
        # next user prompt, where no allocations trigger automatic GC.
        # Cancelled and rescheduled on every _drain_mt_ops call, so it
        # only fires during genuine idle periods.
        if self._gc_on_delete:
            if self._gc_idle_handle is not None:
                self._gc_idle_handle.cancel()
            self._gc_idle_handle = self._loop.call_later(0.1, self._run_idle_gc)

    def _run_idle_gc(self) -> None:
        """Reconcile biased refcounts during idle periods."""
        self._gc_idle_handle = None
        gc.collect(0)

    def _setup_cpp_submit(self) -> None:
        """Set up the C++ raw submit buffer for bypassing Python on the fast path.

        Must be called after the stream is started (self._loop is set).
        The C++ buffer is used only by dispatch_cached_aten's fast path
        (no new tensors). All Python-side submit_execute_aten_bytes calls
        continue through _mt_ops for proven thread-safety.
        """
        if not self._loop:
            return
        try:
            from skytorch.torch.backend._C import _setup_cpp_submit

            _setup_cpp_submit(self._loop.call_soon_threadsafe, self._drain_cpp_raw)
            self._cpp_submit_available = True
        except (ImportError, AttributeError):
            pass

    def _drain_cpp_raw(self) -> None:
        """Drain the C++ raw submit buffer. Must run on the event loop thread.

        Delegates to _drain_mt_ops so that C++ buffer ops are routed through
        _mt_ops, preserving FIFO ordering with "req" entries (update_tensor,
        delete_tensors, etc.). Without this, a C++ op submitted after a "req"
        entry could be drained directly into _raw_batch_buffer, then flushed
        ahead of the "req" by _enqueue_with_flush.
        """
        self._drain_mt_ops()

    def _process_mt_ops(self, ops: list[tuple]) -> None:
        """Process a list of (type, data) ops in order. Must run on the event loop thread."""
        for op_type, data in ops:
            if op_type == "raw":
                self._enqueue_execute_aten_bytes(data)
            elif op_type == "cpp_batch":
                # Pre-batched by C++ flush_raw_batches: (bytes, count).
                # Flush any pending _raw_batch_buffer first to preserve FIFO
                # ordering with prior "raw" entries.
                self._flush_raw_batch()
                batch_bytes, count = data
                if PROFILING_ENABLED:
                    from skytorch.torch.profiler import ClientProfiler

                    _prof = ClientProfiler.get()
                    _prof.batch_count.add_count()
                    _prof.batch_size_total += count
                    if count > _prof.batch_size_max:
                        _prof.batch_size_max = count
                if count == 1:
                    req = service_pb2.StreamRequest(raw_execute_aten=batch_bytes)
                else:
                    req = service_pb2.StreamRequest(raw_batched_execute_aten=batch_bytes)
                self._request_queue.put_nowait(req)
            elif op_type == "aten":
                self._enqueue_execute_aten(data)
            elif op_type == "req":
                self._enqueue_with_flush(data)
            elif op_type == "ff":
                self._deferred_delete_ids.extend(data)
        # Deferred deletes piggyback on the next natural batch flush:
        # _flush_raw_batch calls _send_deferred_deletes after sending the
        # batch, so deletes ride with ATen ops without fragmenting batches.
        # This timer is a safety net for when no ATen ops are pending
        # (e.g., the turn just ended and only deletes remain).
        if self._deferred_delete_ids and not self._deferred_flush_scheduled:
            self._deferred_flush_scheduled = True
            self._deferred_flush_timer = self._loop.call_later(
                self._BATCH_COALESCE_DELAY, self._flush_deferred
            )

    def _flush_mt_ops(self) -> None:
        """Drain main-thread ops buffer at sync points. Must run on the event loop thread."""
        with self._mt_lock:
            # Drain C++ raw submit buffer INTO _mt_ops (not _raw_batch_buffer)
            # to preserve FIFO ordering with "req" entries like update_tensor.
            # Using _drain_cpp_raw() here would put C++ ops into _raw_batch_buffer
            # before _mt_ops is processed, causing _enqueue_with_flush to flush
            # those ops ahead of the "req" entries that should precede them.
            self._drain_cpp_into_mt_ops()
            if not self._mt_ops:
                self._mt_wake_pending = False
                return
            batch = self._mt_ops
            self._mt_ops = []
            self._mt_wake_pending = False
        self._process_mt_ops(batch)

    def _send_deferred_deletes(self) -> None:
        """Send accumulated deferred delete IDs without flushing batch buffers.

        Caller must ensure that any ATen ops referencing the deleted tensors
        have already been sent (either via batch flush or because they were
        enqueued before this call).

        On the free-threaded build, periodically runs gc.collect(0) before
        sending to reconcile biased refcounts from non-owning threads
        (e.g., AsyncStreamer worker).  This is driven by actual delete
        volume, not arbitrary time intervals.

        Must run on the event loop thread.
        """
        self._deferred_flush_scheduled = False
        if self._deferred_flush_timer is not None:
            self._deferred_flush_timer.cancel()
            self._deferred_flush_timer = None
        if not self._deferred_delete_ids:
            return

        # On the free-threaded build, reconcile biased refcounts when live
        # GPU storage grows significantly — indicating biased-refcount
        # tensors are accumulating (allocated but not freed because their
        # deferred decrefs haven't been reconciled).
        if self._gc_on_delete:
            from skytorch.torch.backend._storage import storage_manager

            live = storage_manager.live_storage_bytes
            if live - self._gc_live_bytes_at_last_gc >= self._GC_DELETE_BYTES_THRESHOLD:
                gc.collect(0)
                self._gc_live_bytes_at_last_gc = storage_manager.live_storage_bytes

        ids = self._deferred_delete_ids
        self._deferred_delete_ids = []
        request = service_pb2.DeleteTensorsRequest(tensor_ids=ids)
        stream_request = service_pb2.StreamRequest(delete_tensors=request)
        self._request_queue.put_nowait(stream_request)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Flushed {len(ids)} deferred deletes")

    def _flush_deferred(self) -> None:
        """Flush deferred delete tensor IDs, ensuring batch buffers go first.

        Flushes any pending ATen ops before sending deletes so that ops
        referencing these tensors reach the server first (FIFO ordering).
        This includes draining the C++ raw submit buffer — ops submitted
        before the tensor was freed may still be in the buffer when the
        deferred delete timer fires.

        Must run on the event loop thread.
        """
        # Drain C++ raw submit buffer directly into the request queue.
        # This avoids _drain_mt_ops (which causes re-entrancy through
        # _process_mt_ops → _enqueue_with_flush → _flush_deferred).
        # FIFO ordering between C++ and Python ops is not critical here —
        # the ops referencing deleted tensors are in the C++ buffer, and
        # they must reach the server before the delete message.
        if self._cpp_submit_available:
            try:
                from skytorch.torch.backend._C import _flush_raw_batches

                with self._mt_lock:
                    batches = _flush_raw_batches(self._BATCH_FLUSH_THRESHOLD)
                for batch_bytes, count in batches:
                    if count == 1:
                        req = service_pb2.StreamRequest(raw_execute_aten=batch_bytes)
                    else:
                        req = service_pb2.StreamRequest(
                            raw_batched_execute_aten=batch_bytes
                        )
                    self._request_queue.put_nowait(req)
            except (ImportError, AttributeError):
                pass
        # Flush remaining batch buffers (Python fallback path ops).
        self._flush_batch()
        self._flush_raw_batch()
        self._send_deferred_deletes()

    def defer_delete_ids(self, tensor_ids: list[int]) -> None:
        """Append tensor IDs for deferred batch deletion. Must run on event loop thread.

        Called directly from the GC callback (via call_soon_threadsafe) to bypass
        the ensure_future → coroutine → _mt_ops roundtrip. IDs accumulate until
        the coalesce timer fires, the threshold is hit, or a sync point flushes.
        """
        self._deferred_delete_ids.extend(tensor_ids)
        if len(self._deferred_delete_ids) >= self._DELETE_FLUSH_THRESHOLD:
            self._flush_deferred()
        elif not self._deferred_flush_scheduled:
            self._deferred_flush_scheduled = True
            self._deferred_flush_timer = self._loop.call_later(
                self._DELETE_COALESCE_DELAY, self._flush_deferred
            )

    def _enqueue_with_flush(self, stream_request):
        """Flush any pending batch, then enqueue a request. Must run on the event loop thread."""
        self._flush_batch()
        self._flush_raw_batch()
        self._flush_deferred()
        self._request_queue.put_nowait(stream_request)

    def _submit_request(self, stream_request: service_pb2.StreamRequest, request_type: str) -> None:
        """
        Submit a fire-and-forget request to the stream (callable from any thread).

        Routes through the unified main-thread ops buffer to preserve FIFO
        ordering with submit_execute_aten_bytes calls.

        Args:
            stream_request: StreamRequest to submit
            request_type: Type of request for tracking
        """
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            self._mt_ops.append(("req", stream_request))
            if not self._mt_wake_pending:
                self._mt_wake_pending = True
                self._loop.call_soon_threadsafe(self._drain_mt_ops)

    def submit_execute_aten(self, request: service_pb2.ExecuteAtenRequest) -> None:
        """Submit a fire-and-forget execute_aten request (callable from any thread).

        Requests are buffered and flushed as a batch when the event loop
        processes pending callbacks, reducing per-operation gRPC overhead.
        """
        if logger.isEnabledFor(logging.DEBUG):
            input_tensor_ids = [
                arg.tensor.tensor_id for arg in request.args if arg.WhichOneof("value") == "tensor"
            ]
            output_tensor_ids = [ref.tensor_id for ref in request.outputs]
            logger.debug(
                f"Executing {request.op_name} | "
                f"inputs={input_tensor_ids} | outputs={output_tensor_ids}"
            )
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            self._mt_ops.append(("aten", request))
            if not self._mt_wake_pending:
                self._mt_wake_pending = True
                self._loop.call_soon_threadsafe(self._drain_mt_ops)

    def submit_execute_aten_bytes(self, raw_bytes: bytes) -> None:
        """Submit a fire-and-forget binary-serialized execute_aten request.

        Like submit_execute_aten but takes pre-serialized bytes from
        the C++ request builder, avoiding Python protobuf overhead entirely.
        Callable from any thread.

        Ops are collected in a unified main-thread buffer along with other
        submit calls, preserving FIFO ordering while reducing
        call_soon_threadsafe wake-ups by ~98%.
        """
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            self._mt_ops.append(("raw", raw_bytes))
            if not self._mt_wake_pending:
                self._mt_wake_pending = True
                self._loop.call_soon_threadsafe(self._drain_mt_ops)

    def submit_copy_tensor(self, request: service_pb2.CopyTensorRequest) -> None:
        """Submit a fire-and-forget copy_tensor request (callable from any thread).

        Serialized as binary with a 0xFE marker byte prefix:
        [0xFE][src_id:u64][dst_id:u64][has_src:u8][src_meta...][has_dst:u8][dst_meta...]

        Where metadata (if present) uses the same binary format as ATen op
        tensor metadata, enabling GIL-free server-side parsing.
        """
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Copying tensor {request.src_tensor_id} " f"to tensor {request.dst_tensor_id}"
            )
        parts = [
            struct.pack("<BQQ", 0xFE, request.src_tensor_id, request.dst_tensor_id),
        ]
        for field in ("src_metadata", "dst_metadata"):
            if request.HasField(field):
                meta = getattr(request, field)
                parts.append(b"\x01")
                parts.append(_serialize_tensor_metadata_binary(meta))
            else:
                parts.append(b"\x00")
        self.submit_execute_aten_bytes(b"".join(parts))

    def submit_register_tensors(self, request: service_pb2.RegisterTensorsRequest) -> None:
        """Submit a fire-and-forget register_tensors request (callable from any thread)."""
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Registering {len(request.registrations)} tensors")
        stream_request = service_pb2.StreamRequest(register_tensors=request)
        self._submit_request(stream_request, "register_tensors")

    def submit_delete_tensors(self, request: service_pb2.DeleteTensorsRequest) -> None:
        """Submit a fire-and-forget delete_tensors request (callable from any thread).

        Delete IDs are routed through the deferred "ff" path, accumulating in
        _deferred_delete_ids and batched into a single DeleteTensorsRequest on
        the next flush (batch timer, threshold, or sync point), reducing
        per-message gRPC overhead.
        """
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Deleting tensors {list(request.tensor_ids)}")
        tensor_ids = list(request.tensor_ids)
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            self._mt_ops.append(("ff", tensor_ids))
            if not self._mt_wake_pending:
                self._mt_wake_pending = True
                self._loop.call_soon_threadsafe(self._drain_mt_ops)

    def submit_update_tensor(self, request: service_pb2.UpdateTensorRequest) -> None:
        """Submit a fire-and-forget update_tensor request (callable from any thread)."""
        data_size = len(request.data)

        if logger.isEnabledFor(logging.DEBUG):
            chunked = data_size > STREAM_CHUNK_SIZE
            msg = (
                f"Updating tensor {request.tensor_id} | "
                f"shape={list(request.shape)} | "
                f"size={data_size} bytes"
            )
            if chunked:
                num_chunks = (data_size + STREAM_CHUNK_SIZE - 1) // STREAM_CHUNK_SIZE
                msg += f" | chunks={num_chunks}"
            logger.debug(msg)

        if data_size <= STREAM_CHUNK_SIZE:
            stream_request = service_pb2.StreamRequest(update_tensor=request)
            self._submit_request(stream_request, "update_tensor")
        else:
            self._submit_chunked_update_tensor(request)

    def _submit_chunked_update_tensor(self, request: service_pb2.UpdateTensorRequest) -> None:
        """Submit a large update_tensor request as multiple fire-and-forget chunks."""
        data = request.data
        total_size = len(data)
        total_chunks = (total_size + STREAM_CHUNK_SIZE - 1) // STREAM_CHUNK_SIZE

        mv = memoryview(data)
        stream_requests = []

        for chunk_number in range(total_chunks):
            start = chunk_number * STREAM_CHUNK_SIZE
            end = min(start + STREAM_CHUNK_SIZE, total_size)
            chunk_data = bytes(mv[start:end])

            chunk_request = service_pb2.UpdateTensorRequest(
                tensor_id=request.tensor_id,
                data=chunk_data,
            )
            if chunk_number == 0:
                chunk_request.shape.extend(request.shape)
                chunk_request.dtype = request.dtype
                chunk_request.stride.extend(request.stride)
                chunk_request.storage_offset = request.storage_offset
                if request.HasField("metadata"):
                    chunk_request.metadata.CopyFrom(request.metadata)

            stream_requests.append(
                service_pb2.StreamRequest(
                    update_tensor=chunk_request,
                    chunk_number=chunk_number,
                    total_chunks=total_chunks,
                    total_bytes=total_size if chunk_number == 0 else 0,
                )
            )

        # Route chunks through _mt_ops to preserve FIFO ordering with
        # execute_aten ops. Previously, chunks were enqueued directly on
        # _request_queue via call_soon_threadsafe, but the callback flushed
        # _mt_ops first — sending execute_aten ops that reference the tensor
        # BEFORE the update chunks that create it on the server.
        with self._mt_lock:
            self._drain_cpp_into_mt_ops()
            for request in stream_requests:
                self._mt_ops.append(("req", request))
            if not self._mt_wake_pending:
                self._mt_wake_pending = True
                self._loop.call_soon_threadsafe(self._drain_mt_ops)

    def submit_get_tensor(self, request: service_pb2.GetTensorRequest) -> asyncio.Future:
        """
        Submit a get_tensor request.

        Must be called from the event loop thread.

        Args:
            request: GetTensorRequest to submit

        Returns:
            Awaitable that resolves to StreamResponse with tensor data,
            after checking for deferred errors from prior operations
        """
        if self._loop is None:
            raise RuntimeError("StreamManager not started")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Getting tensor {request.tensor_id}")

        if PROFILING_ENABLED:
            _t_flush_start = time.perf_counter_ns()
            _mt_ops_count = len(self._mt_ops)

        # Flush any pending batches before sync operation.
        # Suppress deferred delete piggyback — deletes must go AFTER
        # the sync request to avoid "Tensor not found" races.
        self._flush_mt_ops()
        self._flush_batch()
        self._flush_raw_batch(piggyback_deletes=False)

        if PROFILING_ENABLED:
            from skytorch.torch.profiler import ClientProfiler

            _t_flush_end = time.perf_counter_ns()
            _prof = ClientProfiler.get()
            _prof.sync_flush.add(_t_flush_end - _t_flush_start)
            _prof.sync_mt_ops_total += _mt_ops_count
            _qdepth = self._request_queue.qsize()
            _prof.sync_queue_depth_total += _qdepth
            if _qdepth > _prof.sync_queue_depth_max:
                _prof.sync_queue_depth_max = _qdepth

        future = self._loop.create_future()
        stream_request = service_pb2.StreamRequest(get_tensor=request)

        self._pending.append(
            PendingRequest(
                future=future,
                request_type="get_tensor",
            )
        )
        self._request_queue.put_nowait(stream_request)

        # Flush deferred deletes AFTER the sync request so the server
        # processes the query before any deletions.
        self._flush_deferred()

        if PROFILING_ENABLED:
            _t_enqueue = time.perf_counter_ns()

        async def _with_error_check():
            try:
                response = await future
            except RuntimeError:
                # If there's a prior deferred error (e.g., OOM from a
                # fire-and-forget op), it's the root cause — raise it
                # instead of the symptom (e.g., "Tensor not found").
                # A speculative get_scalar may have consumed the server's
                # deferred error, leaving subsequent syncs with a
                # confusing secondary error.
                self.check_error()
                raise
            if PROFILING_ENABLED:
                _t_resolved = time.perf_counter_ns()
                _wait_ns = _t_resolved - _t_enqueue
                _prof.sync_wait.add(_wait_ns)
                # Decompose sync_wait using server-provided timing
                _server_total = response.server_backlog_ns + response.server_handle_ns
                _network_rtt = _wait_ns - _server_total
                _prof.sync_network_rtt.add(max(0, _network_rtt))
                _prof.sync_server_backlog.add(response.server_backlog_ns)
                _prof.sync_server_handle.add(response.server_handle_ns)
            # FIFO ordering guarantees all prior fire-and-forget errors are recorded
            self.check_error()
            return response

        return _with_error_check()

    def submit_get_scalar(self, request: service_pb2.GetScalarRequest) -> asyncio.Future:
        """
        Submit a get_scalar request.

        Must be called from the event loop thread.

        Args:
            request: GetScalarRequest to submit

        Returns:
            Awaitable that resolves to StreamResponse with scalar value,
            after checking for deferred errors from prior operations
        """
        if self._loop is None:
            raise RuntimeError("StreamManager not started")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Getting scalar for tensor {request.tensor_id}")

        if PROFILING_ENABLED:
            _t_flush_start = time.perf_counter_ns()
            _mt_ops_count = len(self._mt_ops)

        # Flush any pending batches before sync operation.
        # Suppress deferred delete piggyback — deletes must go AFTER
        # the sync request to avoid "Tensor not found" races.
        self._flush_mt_ops()
        self._flush_batch()
        self._flush_raw_batch(piggyback_deletes=False)

        if PROFILING_ENABLED:
            from skytorch.torch.profiler import ClientProfiler

            _t_flush_end = time.perf_counter_ns()
            _prof = ClientProfiler.get()
            _prof.sync_flush.add(_t_flush_end - _t_flush_start)
            _prof.sync_mt_ops_total += _mt_ops_count
            _qdepth = self._request_queue.qsize()
            _prof.sync_queue_depth_total += _qdepth
            if _qdepth > _prof.sync_queue_depth_max:
                _prof.sync_queue_depth_max = _qdepth

        future = self._loop.create_future()
        stream_request = service_pb2.StreamRequest(get_scalar=request)

        self._pending.append(
            PendingRequest(
                future=future,
                request_type="get_scalar",
            )
        )
        self._request_queue.put_nowait(stream_request)

        # Flush deferred deletes AFTER the sync request so the server
        # processes the query before any deletions.
        self._flush_deferred()

        if PROFILING_ENABLED:
            _t_enqueue = time.perf_counter_ns()

        async def _with_error_check():
            try:
                response = await future
            except RuntimeError:
                # If there's a prior deferred error (e.g., OOM from a
                # fire-and-forget op), it's the root cause — raise it
                # instead of the symptom (e.g., "Tensor not found").
                # A speculative get_scalar may have consumed the server's
                # deferred error, leaving subsequent syncs with a
                # confusing secondary error.
                self.check_error()
                raise
            if PROFILING_ENABLED:
                _t_resolved = time.perf_counter_ns()
                _wait_ns = _t_resolved - _t_enqueue
                _prof.sync_wait.add(_wait_ns)
                # Decompose sync_wait using server-provided timing
                _server_total = response.server_backlog_ns + response.server_handle_ns
                _network_rtt = _wait_ns - _server_total
                _prof.sync_network_rtt.add(max(0, _network_rtt))
                _prof.sync_server_backlog.add(response.server_backlog_ns)
                _prof.sync_server_handle.add(response.server_handle_ns)
            # FIFO ordering guarantees all prior fire-and-forget errors are recorded
            self.check_error()
            return response

        return _with_error_check()

    def submit_execute_module_forward(
        self, request: service_pb2.ExecuteModuleForwardRequest
    ) -> asyncio.Future:
        """
        Submit an execute_module_forward request (sync).

        Must be called from the event loop thread. Flushes pending batches
        before sending to ensure all prior ops complete first.

        Args:
            request: ExecuteModuleForwardRequest to submit

        Returns:
            Awaitable that resolves to StreamResponse with module forward result
        """
        if self._loop is None:
            raise RuntimeError("StreamManager not started")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Executing module forward: model_id={request.model_id} "
                f"module={request.module_path} inputs={list(request.input_tensor_ids)}"
            )

        # Flush any pending batches before sync operation.
        # Suppress deferred delete piggyback — deletes must go AFTER
        # the sync request to avoid "Tensor not found" races.
        self._flush_mt_ops()
        self._flush_batch()
        self._flush_raw_batch(piggyback_deletes=False)

        future = self._loop.create_future()
        stream_request = service_pb2.StreamRequest(execute_module_forward=request)

        self._pending.append(
            PendingRequest(
                future=future,
                request_type="execute_module_forward",
            )
        )
        self._request_queue.put_nowait(stream_request)

        # Flush deferred deletes AFTER the sync request.
        self._flush_deferred()

        async def _with_error_check():
            response = await future
            self.check_error()
            return response

        return _with_error_check()

    def submit_execute_module_forward_ff(
        self, request: service_pb2.ExecuteModuleForwardRequest
    ) -> None:
        """Submit a fire-and-forget execute_module_forward request (callable from any thread).

        Used when output tensors are pre-allocated by the client (cached shapes).
        The request must have output_tensor_ids and output_metadata populated.

        Serialized as binary with a 0xFF marker byte prefix:
        [0xFF][model_id:u64][path_len:u16][path:utf8][n_in:u8][in_ids:u64...][n_out:u8][out_ids:u64...]

        Routed through the raw binary ATen batching pipeline, so module forward
        ops are batched with surrounding ATen ops.
        """
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Executing module forward (ff): model_id={request.model_id} "
                f"module={request.module_path} outputs={list(request.output_tensor_ids)}"
            )
        path_bytes = request.module_path.encode("utf-8")
        n_in = len(request.input_tensor_ids)
        n_out = len(request.output_tensor_ids)
        raw_bytes = struct.pack(
            f"<BQH{len(path_bytes)}sB{n_in}QB{n_out}Q",
            0xFF,
            request.model_id,
            len(path_bytes),
            path_bytes,
            n_in,
            *request.input_tensor_ids,
            n_out,
            *request.output_tensor_ids,
        )
        self.submit_execute_aten_bytes(raw_bytes)

    def submit_release_model(self, request: service_pb2.ReleaseModelRequest) -> None:
        """Submit a fire-and-forget release_model request (callable from any thread)."""
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Releasing model {request.model_id}")
        stream_request = service_pb2.StreamRequest(release_model=request)
        self._submit_request(stream_request, "release_model")

    async def close(self) -> None:
        """Gracefully close the stream."""
        if not self._started:
            return

        # Flush pending batches before closing
        self._flush_mt_ops()
        self._flush_batch()
        self._flush_raw_batch()
        self._flush_deferred()

        # Clear C++ submit buffer state
        if self._cpp_submit_available:
            self._cpp_submit_available = False
            try:
                from skytorch.torch.backend._C import _clear_cpp_submit

                _clear_cpp_submit()
            except (ImportError, AttributeError):
                pass

        self._closing = True

        # Wake up sender loop
        if self._shutdown_event is not None:
            self._shutdown_event.set()

        # Signal sender to stop — the request_generator yields no more requests,
        # the server sees end-of-stream, runs its finally cleanup, and closes
        # the response stream. The receiver completes when it sees the stream end.
        if self._request_queue is not None:
            self._request_queue.put_nowait(None)

        # Wait for sender to finish (it returns once request_generator ends)
        if self._sender_task is not None:
            try:
                await asyncio.wait_for(self._sender_task, timeout=5.0)
            except (asyncio.CancelledError, asyncio.TimeoutError):
                self._sender_task.cancel()
                try:
                    await self._sender_task
                except asyncio.CancelledError:
                    pass

        # Wait for receiver to finish (server closes stream after cleanup)
        if self._receiver_task is not None:
            try:
                await asyncio.wait_for(self._receiver_task, timeout=5.0)
            except (asyncio.CancelledError, asyncio.TimeoutError):
                if self._stream_call is not None:
                    self._stream_call.cancel()
                self._receiver_task.cancel()
                try:
                    await self._receiver_task
                except asyncio.CancelledError:
                    pass

        # Fail any remaining pending requests
        self._fail_all_pending("Stream closed")

        self._started = False

        if PROFILING_ENABLED:
            from skytorch.torch.profiler import ClientProfiler

            ClientProfiler.get().print_summary()
            ClientProfiler.reset()



from openai_harmony import (
    HarmonyEncodingName,
    Role,
    StreamableParser,
    StreamState,
    load_harmony_encoding,
)
from transformers import TextStreamer


class HarmonyStreamer(TextStreamer):
    """Streamer for GPT-OSS Harmony response format.

    Uses openai-harmony's StreamableParser to parse the structured output with
    <|start|>, <|channel|>, <|message|>, <|end|>, and <|return|> special tokens.
    Displays analysis/thinking traces in grey and the final response in normal color.
    """

    _GREY = "\033[90m"
    _RESET = "\033[0m"

    def __init__(self, tokenizer, **kwargs):
        super().__init__(tokenizer, **kwargs)
        self._encoding = load_harmony_encoding(HarmonyEncodingName.HARMONY_GPT_OSS)
        self._start_token_id = self._encoding.encode("<|start|>", allowed_special={"<|start|>"})[0]
        self._endoftext_token_id = self._encoding.encode(
            "<|endoftext|>", allowed_special={"<|endoftext|>"}
        )[0]
        self.reset()

    def reset(self):
        """Clear state between turns."""
        self._parser = StreamableParser(self._encoding, Role.ASSISTANT, strict=False)
        self._prev_state = self._parser.state
        self._prev_channel = None
        self._final_text = []
        self.token_cache = []
        self.print_len = 0
        self.next_tokens_are_prompt = True
        self.generated_tokens = 0
        self._response_complete = False
        self._valid_generated = 0

    def put(self, value):
        if len(value.shape) > 1 and value.shape[0] > 1:
            raise ValueError("HarmonyStreamer only supports batch size 1")
        elif len(value.shape) > 1:
            value = value[0]

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        # Fast path: skip expensive .tolist() (server sync ~150ms) for
        # post-EOS overshoot tokens.  Still count them via shape metadata
        # (local, no sync) so overshoot is computed correctly for KV crop.
        if self._response_complete:
            self.generated_tokens += value.numel()
            return

        for token_id in value.tolist():
            self.generated_tokens += 1
            # <|endoftext|> is never valid content — skip it in any parser state.
            # With strict=False the parser would pass it through as a content delta.
            if token_id == self._endoftext_token_id:
                return
            # After <|end|>, the parser is in EXPECT_START. Only <|start|>
            # is valid (next message in a multi-message response). Skip
            # everything else (<|return|>, trailing tokens, etc.).
            if self._parser.state == StreamState.EXPECT_START:
                if token_id != self._start_token_id:
                    continue
            self._parser.process(token_id)
            state = self._parser.state
            channel = self._parser.current_channel

            # Final response completed (<|end|> → EXPECT_START after "final"
            # channel): record valid token count. Tokens after the analysis
            # channel's <|end|> are NOT overshoot — the final response follows.
            if state == StreamState.EXPECT_START and self._prev_channel == "final":
                self._response_complete = True
                self._valid_generated = self.generated_tokens

            # Transition out of content: reset ANSI for analysis
            if self._prev_state == StreamState.CONTENT and state != StreamState.CONTENT:
                if self._prev_channel == "analysis":
                    print(self._RESET, end="", flush=True)

            # Transition into content: print channel label
            if state == StreamState.CONTENT and self._prev_state != StreamState.CONTENT:
                if channel == "analysis":
                    print(f"\n{self._GREY}Thinking: ", end="", flush=True)
                elif channel == "final":
                    print("\n\nAssistant: ", end="", flush=True)

            # Stream content deltas
            if state == StreamState.CONTENT:
                delta = self._parser.last_content_delta
                if delta:
                    print(delta, end="", flush=True)
                    if channel == "final":
                        self._final_text.append(delta)

            self._prev_state = state
            self._prev_channel = channel

    def end(self):
        if not self._response_complete:
            try:
                self._parser.process_eos()
            except Exception:
                # Response cut off by max_new_tokens before <|end|> —
                # parser is in a non-terminal state (e.g. CONTENT).
                pass
        if self._prev_state == StreamState.CONTENT:
            if self._prev_channel == "analysis":
                print(self._RESET, end="", flush=True)
        self.next_tokens_are_prompt = True
        print()

    @property
    def eos_token_id(self):
        """Harmony stop tokens for use as eos_token_id in model.generate()."""
        return self._encoding.stop_tokens_for_assistant_actions()

    def get_final_response(self):
        """Return text from only the final channel."""
        if not self._final_text:
            return ""
        return "".join(self._final_text).strip()



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from skytorch.client.models.io_k8s_apimachinery_pkg_apis_meta_v1_label_selector_requirement import IoK8sApimachineryPkgApisMetaV1LabelSelectorRequirement
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApimachineryPkgApisMetaV1LabelSelector(BaseModel):
    """
    A label selector is a label query over a set of resources. The result of matchLabels and matchExpressions are ANDed. An empty label selector matches all objects. A null label selector matches no objects.
    """ # noqa: E501
    match_expressions: Optional[List[IoK8sApimachineryPkgApisMetaV1LabelSelectorRequirement]] = Field(default=None, description="matchExpressions is a list of label selector requirements. The requirements are ANDed.", alias="matchExpressions")
    match_labels: Optional[Dict[str, StrictStr]] = Field(default=None, description="matchLabels is a map of {key,value} pairs. A single {key,value} in the matchLabels map is equivalent to an element of matchExpressions, whose key field is \"key\", the operator is \"In\", and the values array contains only \"value\". The requirements are ANDed.", alias="matchLabels")
    __properties: ClassVar[List[str]] = ["matchExpressions", "matchLabels"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1LabelSelector from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in match_expressions (list)
        _items = []
        if self.match_expressions:
            for _item_match_expressions in self.match_expressions:
                if _item_match_expressions:
                    _items.append(_item_match_expressions.to_dict())
            _dict['matchExpressions'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1LabelSelector from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "matchExpressions": [IoK8sApimachineryPkgApisMetaV1LabelSelectorRequirement.from_dict(_item) for _item in obj["matchExpressions"]] if obj.get("matchExpressions") is not None else None,
            "matchLabels": obj.get("matchLabels")
        })
        return _obj





"""
Client for the SkyTorch tensor management gRPC service.

This module provides the TensorClient class with methods for tensor lifecycle
management and ATen operation execution on the remote server.
"""

import logging
from typing import Callable, Optional

try:
    import grpc
    import torch
except ImportError as e:
    raise ImportError(f"Required dependency not found: {e}")

try:
    from skytorch.torch.server import service_pb2
    from skytorch.torch.server import service_pb2_grpc
except ImportError:
    raise ImportError("Generated gRPC code not found. Run hack/gen-grpc-proto.sh first.")

from skytorch.torch.client.tensor import get_tensor_id
from skytorch.torch.client.metadata import TensorMetadata
from skytorch.torch.client.request import (
    tensor_metadata_to_proto,
    build_copy_tensor_request,
    build_execute_aten_request,
)
from skytorch.torch.server.serialization import serialize_tensor_to_chunks, TensorAssembler
from grpc.aio._typing import MetadataType

logger = logging.getLogger(__name__)


class TensorClient:
    """
    Async gRPC client for tensor management and ATen operation execution.

    Provides methods for:
    - Creating tensors on the server
    - Uploading/downloading tensor data
    - Server-side tensor copy
    - Remote ATen operation execution

    Uses a shared gRPC channel provided by the caller (typically GRPCClient).
    """

    def __init__(self, channel: grpc.aio.Channel, metadata: Optional[MetadataType] = None):
        """
        Initialize the client.

        Args:
            channel: gRPC channel to use for communication
            metadata: Optional metadata to include in requests
        """
        self.channel = channel
        self.metadata = metadata
        self.stub = service_pb2_grpc.ServiceStub(self.channel)

    async def update_tensor(
        self,
        src: torch.Tensor,
        tensor_id: int,
        tensor_metadata: Optional[TensorMetadata] = None,
    ) -> None:
        """
        Upload tensor data to server.

        If tensor_metadata is provided, the server will auto-create the tensor
        if it doesn't exist, making this a single operation.

        Args:
            src: Source cpu tensor with data to upload
            tensor_id: Destination tensor ID on the server
            tensor_metadata: Optional metadata for auto-creating the tensor

        Raises:
            RuntimeError: If update fails
        """

        async def stream_tensor():
            first_chunk = True
            for chunk in serialize_tensor_to_chunks(tensor_id, src):
                # Attach sequence number and metadata to first chunk
                if first_chunk:
                    if tensor_metadata is not None:
                        chunk.metadata.CopyFrom(tensor_metadata_to_proto(tensor_metadata))
                    first_chunk = False
                yield chunk

        response = await self.stub.UpdateTensor(stream_tensor(), metadata=self.metadata)

        if not response.success:
            raise RuntimeError(f"Failed to update tensor: {response.message}")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Updated tensor {tensor_id} on server")

    async def get_tensor(
        self,
        tensor_id: int,
        shape: tuple[int, ...],
        dtype: torch.dtype,
        stride: Optional[tuple[int, ...]] = None,
        storage_offset: int = 0,
        tensor_metadata: Optional[TensorMetadata] = None,
    ) -> torch.Tensor:
        """
        Download tensor data from server storage.

        Args:
            tensor_id: Source tensor ID on the server
            shape: Expected tensor shape
            dtype: Expected tensor dtype
            stride: Optional stride (default: contiguous)
            storage_offset: Element offset in the storage
            tensor_metadata: Optional metadata for auto-creating the tensor

        Returns:
            cpu tensor with data from server storage

        Raises:
            RuntimeError: If tensor retrieval fails
        """
        request = service_pb2.GetTensorRequest(
            tensor_id=tensor_id,
            shape=list(shape),
            dtype=str(dtype),
            stride=list(stride) if stride else [],
            storage_offset=storage_offset,
        )
        if tensor_metadata is not None:
            request.metadata.CopyFrom(tensor_metadata_to_proto(tensor_metadata))

        assembler = TensorAssembler()

        async for chunk in self.stub.GetTensor(request, metadata=self.metadata):
            tensor = assembler.add_chunk(chunk)
            if tensor is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        f"Got tensor tensor_id={tensor_id} shape={tensor.shape} "
                        f"data={tensor.flatten()[:8].tolist()}... from server"
                    )
                return tensor

        raise RuntimeError(f"Failed to receive tensor from storage {tensor_id}")

    async def delete_tensors(self, tensor_ids: list[int]) -> None:
        """
        Delete tensors on the server.

        Args:
            tensor_ids: List of tensor IDs to delete

        Raises:
            RuntimeError: If deletion fails
        """
        request = service_pb2.DeleteTensorsRequest(tensor_ids=tensor_ids)
        response = await self.stub.DeleteTensors(request, metadata=self.metadata)

        if not response.success:
            raise RuntimeError(f"Failed to delete tensors: {response.message}")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Deleted {len(tensor_ids)} tensors on server")

    async def copy_tensor(
        self,
        src_tensor_id: int,
        dst_tensor_id: int,
        src_offset: int = 0,
        dst_offset: int = 0,
        num_bytes: int = -1,
        src_metadata: Optional[TensorMetadata] = None,
        dst_metadata: Optional[TensorMetadata] = None,
    ) -> None:
        """
        Copy data between tensors on the server.

        If metadata is provided for source/destination, the server will
        auto-create those tensors if they don't exist, making this atomic.

        Args:
            src_tensor_id: Source tensor ID
            dst_tensor_id: Destination tensor ID
            src_offset: Byte offset in source storage
            dst_offset: Byte offset in destination storage
            num_bytes: Number of bytes to copy (-1 for all)
            src_metadata: Optional metadata for auto-creating source tensor
            dst_metadata: Optional metadata for auto-creating destination tensor

        Raises:
            RuntimeError: If copy fails
        """
        request = build_copy_tensor_request(
            src_tensor_id=src_tensor_id,
            dst_tensor_id=dst_tensor_id,
            src_offset=src_offset,
            dst_offset=dst_offset,
            num_bytes=num_bytes,
            src_metadata=src_metadata,
            dst_metadata=dst_metadata,
        )

        response = await self.stub.CopyTensor(request, metadata=self.metadata)

        if not response.success:
            raise RuntimeError(f"Failed to copy tensor: {response.message}")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Copied tensor {src_tensor_id} -> {dst_tensor_id}")

    async def execute_remote_function(
        self,
        args_bytes: bytes,
        kwargs_bytes: bytes,
        callable_source: str = "",
        callable_name: str = "",
        on_log: Callable[[str, str], None] | None = None,
        retain_model: bool = False,
    ) -> service_pb2.ExecuteFunctionResponse:
        """
        Execute a function on the server.

        Args:
            args_bytes: pickle'd args tuple
            kwargs_bytes: pickle'd kwargs dict
            callable_source: function source code from inspect.getsource()
            callable_name: function name (fn.__name__)
            on_log: Optional callback for log events, called with (stream, text)
                where stream is "stdout" or "stderr"
            retain_model: If True, keep the model alive on the server after
                tensor extraction (for module-level forward proxying)

        Returns:
            ExecuteFunctionResponse with tensor metadata

        Raises:
            RuntimeError: If execution fails
        """
        request = service_pb2.ExecuteFunctionRequest(
            args=args_bytes,
            kwargs=kwargs_bytes,
            callable_source=callable_source,
            callable_name=callable_name,
            retain_model=retain_model,
        )

        response: service_pb2.ExecuteFunctionResponse | None = None
        async for event in self.stub.ExecuteFunction(request, metadata=self.metadata):
            which = event.WhichOneof("event")
            if which == "log":
                if on_log is not None:
                    on_log(event.log.stream, event.log.text)
            elif which == "result":
                response = event.result

        if response is None:
            raise RuntimeError("ExecuteFunction stream ended without a result event")
        return response

    async def execute_aten_operation(
        self,
        op_name: str,
        args: tuple,
        kwargs: dict,
        output_tensors: list[torch.Tensor] | None,
        tensor_metadata: Optional[list[TensorMetadata]] = None,
        output_metadata: Optional[list[TensorMetadata]] = None,
    ) -> list[int] | None:
        """
        Execute an ATen operation on the server.

        Supports two modes:
        - Pre-allocated outputs: output_tensors provided, writes to them, returns None
        - Server-created outputs: output_tensors is None, returns list[int] (tensor_ids)

        If metadata lists are provided, the server will auto-create those tensors
        if they don't exist, making this a single atomic operation.

        Args:
            op_name: ATen operation name (e.g., "aten::add.Tensor")
            args: Positional arguments (may contain sky tensors)
            kwargs: Keyword arguments (may contain sky tensors)
            output_tensors: Pre-allocated output tensors, or None for server-created
            tensor_metadata: Optional list of input tensor metadata for auto-creation
            output_metadata: Optional list of output tensor metadata for auto-creation

        Returns:
            None if output_tensors provided, list[int] of tensor_ids if server created outputs

        Raises:
            RuntimeError: If operation execution fails
        """
        request = build_execute_aten_request(
            op_name=op_name,
            args=args,
            kwargs=kwargs,
            output_tensors=output_tensors,
            tensor_metadata=tensor_metadata,
            output_metadata=output_metadata,
        )

        if logger.isEnabledFor(logging.DEBUG):
            input_tensor_ids = [
                get_tensor_id(arg)
                for arg in args
                if isinstance(arg, torch.Tensor) and arg.device.type == "sky"
            ]
            output_tensor_ids = [get_tensor_id(t) for t in (output_tensors or []) if t is not None]
            logger.debug(
                f"Executing {op_name} | " f"inputs={input_tensor_ids} | outputs={output_tensor_ids}"
            )

        response = await self.stub.ExecuteAtenOperation(request, metadata=self.metadata)

        if not response.success:
            raise RuntimeError(f"ATen operation failed: {response.message}")

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Executed {op_name}")

        if output_tensors is None and response.output_tensors:
            return [ref.tensor_id for ref in response.output_tensors]

        return None



"""Utilities for capturing output from remotely executed functions."""

import io
import logging
import os
import queue


_TEE_FUNCTION_OUTPUT = os.environ.get("SKYTORCH_TEE_FUNCTION_OUTPUT", "false").lower() in (
    "1",
    "true",
    "yes",
)


class OutputCapture(io.TextIOBase):
    """Captures writes to stdout/stderr and puts them onto a queue.

    When ``tee=True``, also writes to the original stream so server-side logs
    are preserved.  Controlled by the ``SKYTORCH_TEE_FUNCTION_OUTPUT`` env var
    (default: false, since output is streamed to the client).
    """

    def __init__(
        self,
        stream_name: str,
        original: io.TextIOBase,
        queue: queue.SimpleQueue,
        tee: bool = _TEE_FUNCTION_OUTPUT,
    ):
        self._stream_name = stream_name
        self._original = original
        self._queue = queue
        self._tee = tee

    def write(self, s: str) -> int:
        if s:
            self._queue.put((self._stream_name, s))
            if self._tee:
                self._original.write(s)
        return len(s)

    def flush(self):
        self._original.flush()

    def fileno(self):
        return self._original.fileno()

    @property
    def encoding(self):
        return self._original.encoding

    def isatty(self):
        return False


class LogCapture(logging.Handler):
    """Logging handler that captures log records onto a queue.

    This intercepts output from libraries (e.g. httpx, transformers) that log
    via Python's logging module rather than printing to stdout/stderr.

    When ``tee=False``, the existing handlers on the root logger are temporarily
    disabled so that log records are only forwarded to the client.
    """

    def __init__(self, log_queue: queue.SimpleQueue, tee: bool = _TEE_FUNCTION_OUTPUT):
        super().__init__()
        self._queue = log_queue
        self._tee = tee
        self._muted_handlers: list[logging.Handler] = []

    def install(self) -> None:
        """Attach to the root logger, optionally muting existing handlers."""
        root = logging.getLogger()
        if not self._tee:
            for h in root.handlers:
                self._muted_handlers.append(h)
                h.addFilter(_reject_all)
        root.addHandler(self)

    def uninstall(self) -> None:
        """Remove from the root logger and restore muted handlers."""
        root = logging.getLogger()
        root.removeHandler(self)
        for h in self._muted_handlers:
            h.removeFilter(_reject_all)
        self._muted_handlers.clear()

    def emit(self, record: logging.LogRecord):
        try:
            msg = self.format(record) + "\n"
            self._queue.put(("stderr", msg))
        except Exception:
            self.handleError(record)


def _reject_all(record: logging.LogRecord) -> bool:
    """Filter that rejects all log records."""
    return False



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MetricType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN: _ClassVar[MetricType]
    GAUGE: _ClassVar[MetricType]
    COUNTER: _ClassVar[MetricType]
    HISTOGRAM: _ClassVar[MetricType]
UNKNOWN: MetricType
GAUGE: MetricType
COUNTER: MetricType
HISTOGRAM: MetricType

class Metric(_message.Message):
    __slots__ = ("name", "type", "value", "unit", "labels", "timestamp", "help")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    UNIT_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    HELP_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: MetricType
    value: float
    unit: str
    labels: _containers.ScalarMap[str, str]
    timestamp: int
    help: str
    def __init__(self, name: _Optional[str] = ..., type: _Optional[_Union[MetricType, str]] = ..., value: _Optional[float] = ..., unit: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ..., timestamp: _Optional[int] = ..., help: _Optional[str] = ...) -> None: ...

class MetricsSnapshot(_message.Message):
    __slots__ = ("metrics", "source", "timestamp")
    METRICS_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    metrics: _containers.RepeatedCompositeFieldContainer[Metric]
    source: str
    timestamp: int
    def __init__(self, metrics: _Optional[_Iterable[_Union[Metric, _Mapping]]] = ..., source: _Optional[str] = ..., timestamp: _Optional[int] = ...) -> None: ...

class GetMetricsRequest(_message.Message):
    __slots__ = ("metric_names", "sources", "label_filters")
    class LabelFiltersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    METRIC_NAMES_FIELD_NUMBER: _ClassVar[int]
    SOURCES_FIELD_NUMBER: _ClassVar[int]
    LABEL_FILTERS_FIELD_NUMBER: _ClassVar[int]
    metric_names: _containers.RepeatedScalarFieldContainer[str]
    sources: _containers.RepeatedScalarFieldContainer[str]
    label_filters: _containers.ScalarMap[str, str]
    def __init__(self, metric_names: _Optional[_Iterable[str]] = ..., sources: _Optional[_Iterable[str]] = ..., label_filters: _Optional[_Mapping[str, str]] = ...) -> None: ...

class GetMetricsResponse(_message.Message):
    __slots__ = ("snapshots",)
    SNAPSHOTS_FIELD_NUMBER: _ClassVar[int]
    snapshots: _containers.RepeatedCompositeFieldContainer[MetricsSnapshot]
    def __init__(self, snapshots: _Optional[_Iterable[_Union[MetricsSnapshot, _Mapping]]] = ...) -> None: ...

class StreamMetricsRequest(_message.Message):
    __slots__ = ("interval_seconds", "metric_names", "sources", "label_filters")
    class LabelFiltersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    METRIC_NAMES_FIELD_NUMBER: _ClassVar[int]
    SOURCES_FIELD_NUMBER: _ClassVar[int]
    LABEL_FILTERS_FIELD_NUMBER: _ClassVar[int]
    interval_seconds: float
    metric_names: _containers.RepeatedScalarFieldContainer[str]
    sources: _containers.RepeatedScalarFieldContainer[str]
    label_filters: _containers.ScalarMap[str, str]
    def __init__(self, interval_seconds: _Optional[float] = ..., metric_names: _Optional[_Iterable[str]] = ..., sources: _Optional[_Iterable[str]] = ..., label_filters: _Optional[_Mapping[str, str]] = ...) -> None: ...



"""
Main entry point for the PyTorch tensor streaming gRPC server.
"""

import argparse
import asyncio
import logging
import os
import signal

import grpc

from skytorch.torch.server.server import serve, graceful_shutdown
from skytorch.torch.server.serialization import DEFAULT_CHUNK_SIZE
from skytorch.server.metrics.source import load_metrics_source, list_metrics_source_names


parser = argparse.ArgumentParser(
    description='SkyTorch Server',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter
)

parser.add_argument(
    '--port',
    type=int,
    default=int(os.environ.get('SKYTORCH_PORT', 50051)),
    help='Port to listen on'
)

parser.add_argument(
    '--host',
    type=str,
    default=os.environ.get('SKYTORCH_HOST', '[::]'),
    help='Host address to bind to (use [::] for all interfaces)'
)

parser.add_argument(
    '--chunk-size',
    type=int,
    default=int(os.environ.get('SKYTORCH_CHUNK_SIZE', DEFAULT_CHUNK_SIZE)),
    help='Size of chunks for streaming tensors (in bytes)'
)

parser.add_argument(
    '--log-level',
    type=str,
    default=os.environ.get('SKYTORCH_LOG_LEVEL', 'INFO'),
    choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    help='Logging level'
)

parser.add_argument(
    '--metrics-sources',
    type=str,
    nargs='*',
    default=os.environ.get('SKYTORCH_METRICS_SOURCES', '').split(',') if os.environ.get('SKYTORCH_METRICS_SOURCES', '') else [],
    help=f'Metrics sources to enable (available: {", ".join(list_metrics_source_names())})'
)

args = parser.parse_args()

logging.basicConfig(
    level=getattr(logging, args.log_level),
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
)

logger = logging.getLogger(__name__)
logger.info(f"Starting SkyTorch Server")
logger.info(f"  Port: {args.port}")
logger.info(f"  Host: {args.host}")
logger.info(f"  Chunk Size: {args.chunk_size} bytes ({args.chunk_size / 1024 / 1024:.2f} MB)")
logger.info(f"  Log Level: {args.log_level}")
logger.info(f"  Metrics Sources: {', '.join(args.metrics_sources) if args.metrics_sources else 'none'}")

_COMPRESSION_MAP = {
    "none": grpc.Compression.NoCompression,
    "deflate": grpc.Compression.Deflate,
    "gzip": grpc.Compression.Gzip,
}
_compression_name = os.environ.get("SKYTORCH_GRPC_COMPRESSION", "none").lower()
_compression = _COMPRESSION_MAP.get(_compression_name, grpc.Compression.Gzip)
logger.info(f"  Compression: {_compression_name}")

loop = asyncio.new_event_loop()
loop.set_task_factory(asyncio.eager_task_factory)
asyncio.set_event_loop(loop)
server = grpc.aio.server(
    compression=_compression,
    options=[
        ("grpc.http2.initial_window_size", 1 * 1024 * 1024),              # 1MB (default 64KB)
        ("grpc.http2.initial_connection_window_size", 2 * 1024 * 1024),   # 2MB
        ("grpc.max_send_message_length", 16 * 1024 * 1024),               # 16MB
        ("grpc.max_receive_message_length", 16 * 1024 * 1024),            # 16MB
        ("grpc.keepalive_time_ms", 30_000),                                # Send keepalive ping every 30s
        ("grpc.keepalive_timeout_ms", 10_000),                             # Wait 10s for ping ack
        ("grpc.keepalive_permit_without_calls", 1),                        # Allow pings without active RPCs
        ("grpc.http2.max_pings_without_data", 0),                          # Allow unlimited pings without data
        ("grpc.http2.min_recv_ping_interval_without_data_ms", 10_000),     # Accept client pings at 10s+
    ],
)

# Initialize metrics sources based on CLI configuration
metrics_sources = []
for source_name in args.metrics_sources:
    source = load_metrics_source(source_name)
    if source is not None:
        if source.is_available():
            metrics_sources.append(source)
            logger.info(f"Enabled metrics source: {source_name}")
        else:
            logger.warning(f"Metrics source '{source_name}' is not available")

for sig in (signal.SIGINT, signal.SIGTERM):
    loop.add_signal_handler(sig, lambda: asyncio.ensure_future(graceful_shutdown(server, metrics_sources, grace=5.0)))

try:
    loop.run_until_complete(
        serve(
            server,
            host=args.host,
            port=args.port,
            chunk_size=args.chunk_size,
            metrics_sources=metrics_sources,
        )
    )
finally:
    loop.run_until_complete(server.wait_for_termination())
    loop.close()



"""
Async streamer wrapper for HuggingFace Transformers.

Offloads the streamer's put() to a background thread so the generate
loop isn't blocked by DeferredScalarTensor resolution (tolist() calls).
"""

import queue
import threading

import torch
from transformers.generation.streamers import BaseStreamer


class AsyncStreamer(BaseStreamer):
    """Wraps a streamer to offload put() to a background thread.

    This prevents the generate thread from blocking on DeferredScalarTensor
    resolution (tolist() inside the inner streamer's put()). The generate
    loop sees put() return immediately, allowing it to dispatch the next
    forward pass while the streamer processes tokens in parallel.

    Args:
        inner: The streamer to wrap (e.g., TextStreamer, TextIteratorStreamer).
            Must implement put(), end(), and optionally reset().
    """

    _SENTINEL = object()

    def __init__(self, inner: BaseStreamer):
        self._inner = inner
        self._queue: queue.SimpleQueue = queue.SimpleQueue()
        self._error: Exception | None = None
        self._thread = threading.Thread(target=self._worker, daemon=True)
        self._thread.start()

    def _worker(self) -> None:
        try:
            while True:
                item = self._queue.get()
                if item is self._SENTINEL:
                    break
                self._inner.put(item)
        except Exception as e:
            self._error = e

    def put(self, value: torch.Tensor) -> None:
        if self._error is not None:
            raise self._error
        self._queue.put(value)

    def reset(self) -> None:
        if self._thread.is_alive():
            self._queue.put(self._SENTINEL)
            self._thread.join()
        self._queue = queue.SimpleQueue()
        self._error = None
        self._thread = threading.Thread(target=self._worker, daemon=True)
        self._thread.start()
        self._inner.reset()

    def end(self) -> None:
        self._queue.put(self._SENTINEL)
        self._thread.join()
        if self._error is not None:
            raise self._error
        self._inner.end()

    def __getattr__(self, name: str):
        return getattr(self._inner, name)



import logging
from os.path import exists, expanduser
from typing import Optional

from kubernetes.client import Configuration
from kubernetes.config import KUBE_CONFIG_DEFAULT_LOCATION, ConfigException
from kubernetes.config import (
    list_kube_config_contexts,
    load_kube_config,
    load_incluster_config,
)

logger = logging.getLogger(__name__)


# Default Kubernetes namespace
_default_namespace = "default"


def default_namespace() -> str:
    """
    Get the default namespace.

    Returns the namespace that will be used for Compute resources. This value
    is set either explicitly via init(namespace=...) or auto-detected from
    kubeconfig context or service account.

    If init() has not been called yet, this function will automatically call it
    to trigger namespace auto-detection. The namespace is auto-detected by reading
    from kubeconfig context or the service account namespace file.

    Returns:
        The default namespace (auto-detected or explicitly set via init())
    """
    # Auto-initialize if not already done
    if Configuration._default is None:
        init()

    return _default_namespace


def init(
        *,
        client_config: Optional[Configuration] = None,
        namespace: Optional[str] = None,
):
    """
    Initialize Kubernetes client configuration globally.

    This is optional. If not called, the client will auto-initialize on first use
    by trying default kubeconfig first, then falling back to in-cluster config.

    The default namespace is also auto-detected (unless explicitly provided):
    - From kubeconfig: Uses the namespace from the current context
    - In-cluster: Reads from /var/run/secrets/kubernetes.io/serviceaccount/namespace
    - Falls back to "default" if no namespace is detected

    Args:
        client_config: Kubernetes client Configuration object. Use this for advanced
                      configuration like custom timeouts, SSL settings, or API server URL.
        namespace: Namespace to use for all Compute resources. If not provided, it will
                  be auto-detected from kubeconfig context or service account.

    Raises:
        RuntimeError: If the client has already been initialized.

    Example:
        >>> from kubernetes import client
        >>> from skytorch.client import init
        >>>
        >>> # Advanced configuration with custom namespace
        >>> config = client.Configuration()
        >>> config.host = "https://my-k8s-cluster.example.com"
        >>> config.api_key_prefix['authorization'] = 'Bearer'
        >>> config.api_key['authorization'] = 'my-token'
        >>> init(client_config=config, namespace="my-namespace")
        >>>
        >>> # Or just override the namespace
        >>> init(namespace="my-namespace")
        >>>
        >>> # Or let it auto-detect (default kubeconfig -> in-cluster)
        >>> # No need to call init() at all, it will happen lazily
    """
    global _default_namespace

    if Configuration._default is not None:
        raise RuntimeError(
            "Kubernetes client is already initialized. "
            "The init() function can only be called once before creating any Compute instances."
        )

    # If namespace is explicitly provided, use it and skip auto-detection
    if namespace is not None:
        _default_namespace = namespace
        logger.debug(f"Using explicitly provided namespace '{_default_namespace}'")

    if client_config is not None:
        # Use provided configuration
        Configuration.set_default(client_config)
        logger.debug("Kubernetes client initialized with provided configuration")
        return

    # Auto-detect: try default kubeconfig first, then fall back to in-cluster
    if exists(expanduser(KUBE_CONFIG_DEFAULT_LOCATION)):
        # Load kubeconfig and get the current context
        if namespace is None:
            contexts, active_context = list_kube_config_contexts()

            # Extract namespace from active context
            if active_context and "context" in active_context:
                context_namespace = active_context["context"].get("namespace")
                if context_namespace:
                    _default_namespace = context_namespace
                    logger.debug(f"Using namespace '{_default_namespace}' from kubeconfig context")

        load_kube_config()
        logger.debug("Kubernetes client initialized with default kubeconfig")
        return
    else:
        logger.debug("Default kubeconfig not found, trying in-cluster config")
        try:
            load_incluster_config()
        except ConfigException as e:
            logger.debug(f"In-cluster configuration failed: {e}")
        else:
            if namespace is None:
                namespace_file = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
                if not exists(namespace_file):
                    raise ConfigException(
                        f"Running in-cluster but service account namespace file not found: "
                        f"{namespace_file}"
                    )

                with open(namespace_file, 'r') as f:
                    ns = f.read().strip()
                    if not ns:
                        raise ConfigException(
                            f"Service account namespace file is empty: {namespace_file}"
                        )
                    _default_namespace = ns
                    logger.debug(f"Using namespace '{_default_namespace}' from service account")

            logger.debug("Kubernetes client initialized with in-cluster config")
            return

    # Initialization failed
    raise ConfigException(
        "Could not configure Kubernetes client. "
        f"Neither default kubeconfig ({KUBE_CONFIG_DEFAULT_LOCATION}) "
        "nor in-cluster config could be loaded."
    )
